from typing import Callable

from rx import operators as ops
from rx.core import Observable, pipe


def _buffer_with_time_or_count(timespan, count, scheduler = None) -> Callable[[Observable], Observable]:
    return pipe(
        ops.window_with_time_or_count(timespan, count, scheduler),
        ops.flat_map(lambda x: x.pipe(ops.to_iterable()))
    )



from sys import maxsize
from typing import Optional

from rx.core import typing
from rx.core import Observable
from rx.scheduler import CurrentThreadScheduler
from rx.disposable import MultipleAssignmentDisposable


def _range(start: int,
           stop: Optional[int] = None,
           step: Optional[int] = None,
           scheduler: Optional[typing.Scheduler] = None
           ) -> Observable:
    """Generates an observable sequence of integral numbers within a
    specified range, using the specified scheduler to send out observer
    messages.

    Examples:
        >>> res = range(10)
        >>> res = range(0, 10)
        >>> res = range(0, 10, 1)

    Args:
        start: The value of the first integer in the sequence.
        count: The number of sequential integers to generate.
        scheduler: The scheduler to schedule the values on.

    Returns:
        An observable sequence that contains a range of sequential
        integral numbers.
    """

    _stop: int = maxsize if stop is None else stop
    _step: int = 1 if step is None else step

    if step is None and stop is None:
        range_t = range(start)
    elif step is None:
        range_t = range(start, _stop)
    else:
        range_t = range(start, _stop, _step)

    def subscribe(observer, scheduler_: typing.Scheduler = None):
        nonlocal range_t

        _scheduler = scheduler or scheduler_ or CurrentThreadScheduler.singleton()
        sd = MultipleAssignmentDisposable()

        def action(scheduler, iterator):
            try:
                observer.on_next(next(iterator))
                sd.disposable = _scheduler.schedule(action, state=iterator)
            except StopIteration:
                observer.on_completed()

        sd.disposable = _scheduler.schedule(action, iter(range_t))
        return sd
    return Observable(subscribe)



from typing import Generator, List

import twitch.tmi as tmi
from twitch.api import API
from twitch.baseresource import BaseResource


class Chatters(BaseResource['tmi.Chatter']):

    def __init__(self, api: API, user: str):
        super().__init__(api=api, path='group/user/{user}/chatters')
        self._api = api

        # API return data
        self._data = self._api.get(self._path.format(user=user))

        # API Data
        self.count: int = self._data.get('chatter_count', -1)

        self.types: List[str] = list(self._data.get('chatters', {}).keys())

        self.broadcaster: List[tmi.Chatter] = [tmi.Chatter(self._api, name, 'broadcaster') for name in
                                               self._data.get('chatters', {}).get('broadcaster', [])]

        self.vips: List[tmi.Chatter] = [tmi.Chatter(self._api, name, 'vip') for name in
                                        self._data.get('chatters', {}).get('vips', [])]

        self.moderators: List[tmi.Chatter] = [tmi.Chatter(self._api, name, 'moderator') for name in
                                              self._data.get('chatters', {}).get('moderators', [])]

        self.staff: List[tmi.Chatter] = [tmi.Chatter(self._api, name, 'staff') for name in
                                         self._data.get('chatters', {}).get('staff', [])]

        self.admins: List[tmi.Chatter] = [tmi.Chatter(self._api, name, 'admin') for name in
                                          self._data.get('chatters', {}).get('admins', [])]

        self.global_mods: List[tmi.Chatter] = [tmi.Chatter(self._api, name, 'global_mod') for name in
                                               self._data.get('chatters', {}).get('global_mods', [])]

        self.viewers: List[tmi.Chatter] = [tmi.Chatter(self._api, name, 'viewer') for name in
                                           self._data.get('chatters', {}).get('viewers', [])]

    def all(self) -> List[tmi.Chatter]:
        """
        Get all chatters from all groups
        :return: List of all chatters
        """
        return self.broadcaster + self.vips + self.moderators + self.staff + self.admins + self.global_mods + self.viewers

    def __iter__(self) -> Generator['tmi.Chatter', None, None]:
        """
        Iterate over all chatters
        :return: Yield chatter
        """
        for chatter in self.all():
            yield chatter

    def __getitem__(self, index: int) -> 'tmi.Chatter':
        """
        Get chatter by index
        :param index: Index
        :return: Chatter
        """
        return self.all()[index]



from typing import Callable, NamedTuple, Any, Optional
from datetime import timedelta


from rx import operators as ops
from rx.core import Observable, typing
from rx.scheduler import TimeoutScheduler


class TimeInterval(NamedTuple):
    value: Any
    interval: timedelta


def _time_interval(scheduler: Optional[typing.Scheduler] = None) -> Callable[[Observable], Observable]:
    def time_interval(source: Observable) -> Observable:
        """Records the time interval between consecutive values in an
        observable sequence.

            >>> res = time_interval(source)

        Return:
            An observable sequence with time interval information on
            values.
        """

        def subscribe(observer, scheduler_):
            _scheduler = scheduler or scheduler_ or TimeoutScheduler.singleton()
            last = _scheduler.now

            def mapper(value):
                nonlocal last

                now = _scheduler.now
                span = now - last
                last = now
                return TimeInterval(value=value, interval=span)

            return source.pipe(ops.map(mapper)).subscribe(observer, scheduler_)
        return Observable(subscribe)
    return time_interval



from typing import Any, Callable, Optional

from rx import operators as ops
from rx.core import Observable, pipe
from rx.core.typing import Predicate
from rx.internal.exceptions import SequenceContainsNoElementsError


def _first_or_default_async(has_default=False, default_value=None):
    def first_or_default_async(source: Observable) -> Observable:
        def subscribe(observer, scheduler=None):
            def on_next(x):
                observer.on_next(x)
                observer.on_completed()

            def on_completed():
                if not has_default:
                    observer.on_error(SequenceContainsNoElementsError())
                else:
                    observer.on_next(default_value)
                    observer.on_completed()

            return source.subscribe_(on_next, observer.on_error, on_completed, scheduler)
        return Observable(subscribe)
    return first_or_default_async


def _first_or_default(predicate: Optional[Predicate] = None,
                      default_value: Any = None
                      ) -> Callable[[Observable], Observable]:
    """Returns the first element of an observable sequence that
    satisfies the condition in the predicate, or a default value if no
    such element exists.

    Examples:
        >>> res = source.first_or_default()
        >>> res = source.first_or_default(lambda x: x > 3)
        >>> res = source.first_or_default(lambda x: x > 3, 0)
        >>> res = source.first_or_default(None, 0)

    Args:
        source -- Observable sequence.
        predicate -- [optional] A predicate function to evaluate for
            elements in the source sequence.
        default_value -- [Optional] The default value if no such element
            exists.  If not specified, defaults to None.

    Returns:
        A function that takes an observable source and reutrn an
        observable sequence containing the first element in the
        observable sequence that satisfies the condition in the
        predicate, or a default value if no such element exists.
    """

    if predicate:
        return pipe(
            ops.filter(predicate),
            ops.first_or_default(None, default_value)
        )
    return _first_or_default_async(True, default_value)



from typing import Optional, Callable

from rx import operators as ops
from rx.core import Observable, pipe
from rx.core.typing import Predicate, Any
from rx.internal.exceptions import SequenceContainsNoElementsError


def _single_or_default_async(has_default: bool = False, default_value: Any = None) -> Callable[[Observable], Observable]:
    def single_or_default_async(source: Observable) -> Observable:
        def subscribe(observer, scheduler=None):
            value = [default_value]
            seen_value = [False]

            def on_next(x):
                if seen_value[0]:
                    observer.on_error(Exception('Sequence contains more than one element'))
                else:
                    value[0] = x
                    seen_value[0] = True

            def on_completed():
                if not seen_value[0] and not has_default:
                    observer.on_error(SequenceContainsNoElementsError())
                else:
                    observer.on_next(value[0])
                    observer.on_completed()

            return source.subscribe_(on_next, observer.on_error, on_completed, scheduler)
        return Observable(subscribe)
    return single_or_default_async


def _single_or_default(predicate: Optional[Predicate] = None, default_value: Any = None) -> Callable[[Observable], Observable]:
    """Returns the only element of an observable sequence that matches
    the predicate, or a default value if no such element exists this
    method reports an exception if there is more than one element in the
    observable sequence.

    Examples:
        >>> res = single_or_default()
        >>> res = single_or_default(lambda x: x == 42)
        >>> res = single_or_default(lambda x: x == 42, 0)
        >>> res = single_or_default(None, 0)

    Args:
        predicate -- [Optional] A predicate function to evaluate for
            elements in the source sequence.
        default_value -- [Optional] The default value if the index is
            outside the bounds of the source sequence.

    Returns:
        An observable Sequence containing the single element in the
    observable sequence that satisfies the condition in the predicate,
    or a default value if no such element exists.
    """

    if predicate:
        return pipe(ops.filter(predicate), ops.single_or_default(None, default_value))
    else:
        return _single_or_default_async(True, default_value)



import threading

from rx.core import typing


class InnerSubscription(typing.Disposable):
    def __init__(self, subject, observer):
        self.subject = subject
        self.observer = observer

        self.lock = threading.RLock()

    def dispose(self) -> None:
        with self.lock:
            if not self.subject.is_disposed and self.observer:
                if self.observer in self.subject.observers:
                    self.subject.observers.remove(self.observer)
                self.observer = None



from typing import Callable
from rx.core import Observable


def _to_set() -> Callable[[Observable], Observable]:
    """Converts the observable sequence to a set.

    Returns an observable sequence with a single value of a set
    containing the values from the observable sequence.
    """

    def to_set(source: Observable) -> Observable:
        def subscribe(observer, scheduler=None):
            s = set()

            def on_completed():
                nonlocal s
                observer.on_next(s)
                s = set()
                observer.on_completed()

            return source.subscribe_(s.add, observer.on_error, on_completed, scheduler)
        return Observable(subscribe)
    return to_set


import functools

from matplotlib import _api, docstring
import matplotlib.artist as martist
from matplotlib.axes._axes import Axes
from matplotlib.gridspec import GridSpec, SubplotSpec


class SubplotBase:
    """
    Base class for subplots, which are :class:`Axes` instances with
    additional methods to facilitate generating and manipulating a set
    of :class:`Axes` within a figure.
    """

    def __init__(self, fig, *args, **kwargs):
        """
        Parameters
        ----------
        fig : `matplotlib.figure.Figure`

        *args : tuple (*nrows*, *ncols*, *index*) or int
            The array of subplots in the figure has dimensions ``(nrows,
            ncols)``, and *index* is the index of the subplot being created.
            *index* starts at 1 in the upper left corner and increases to the
            right.

            If *nrows*, *ncols*, and *index* are all single digit numbers, then
            *args* can be passed as a single 3-digit number (e.g. 234 for
            (2, 3, 4)).

        **kwargs
            Keyword arguments are passed to the Axes (sub)class constructor.
        """
        # _axes_class is set in the subplot_class_factory
        self._axes_class.__init__(self, fig, [0, 0, 1, 1], **kwargs)
        # This will also update the axes position.
        self.set_subplotspec(SubplotSpec._from_subplot_args(fig, args))

    def __reduce__(self):
        # get the first axes class which does not inherit from a subplotbase
        axes_class = next(
            c for c in type(self).__mro__
            if issubclass(c, Axes) and not issubclass(c, SubplotBase))
        return (_picklable_subplot_class_constructor,
                (axes_class,),
                self.__getstate__())

    @_api.deprecated(
        "3.4", alternative="get_subplotspec",
        addendum="(get_subplotspec returns a SubplotSpec instance.)")
    def get_geometry(self):
        """Get the subplot geometry, e.g., (2, 2, 3)."""
        rows, cols, num1, num2 = self.get_subplotspec().get_geometry()
        return rows, cols, num1 + 1  # for compatibility

    @_api.deprecated("3.4", alternative="set_subplotspec")
    def change_geometry(self, numrows, numcols, num):
        """Change subplot geometry, e.g., from (1, 1, 1) to (2, 2, 3)."""
        self._subplotspec = GridSpec(numrows, numcols,
                                     figure=self.figure)[num - 1]
        self.update_params()
        self.set_position(self.figbox)

    def get_subplotspec(self):
        """Return the `.SubplotSpec` instance associated with the subplot."""
        return self._subplotspec

    def set_subplotspec(self, subplotspec):
        """Set the `.SubplotSpec`. instance associated with the subplot."""
        self._subplotspec = subplotspec
        self._set_position(subplotspec.get_position(self.figure))

    def get_gridspec(self):
        """Return the `.GridSpec` instance associated with the subplot."""
        return self._subplotspec.get_gridspec()

    @_api.deprecated(
        "3.4", alternative="get_subplotspec().get_position(self.figure)")
    @property
    def figbox(self):
        return self.get_subplotspec().get_position(self.figure)

    @_api.deprecated("3.4", alternative="get_gridspec().nrows")
    @property
    def numRows(self):
        return self.get_gridspec().nrows

    @_api.deprecated("3.4", alternative="get_gridspec().ncols")
    @property
    def numCols(self):
        return self.get_gridspec().ncols

    @_api.deprecated("3.4")
    def update_params(self):
        """Update the subplot position from ``self.figure.subplotpars``."""
        # Now a no-op, as figbox/numRows/numCols are (deprecated) auto-updating
        # properties.

    @_api.deprecated("3.4", alternative="ax.get_subplotspec().is_first_row()")
    def is_first_row(self):
        return self.get_subplotspec().rowspan.start == 0

    @_api.deprecated("3.4", alternative="ax.get_subplotspec().is_last_row()")
    def is_last_row(self):
        return self.get_subplotspec().rowspan.stop == self.get_gridspec().nrows

    @_api.deprecated("3.4", alternative="ax.get_subplotspec().is_first_col()")
    def is_first_col(self):
        return self.get_subplotspec().colspan.start == 0

    @_api.deprecated("3.4", alternative="ax.get_subplotspec().is_last_col()")
    def is_last_col(self):
        return self.get_subplotspec().colspan.stop == self.get_gridspec().ncols

    def label_outer(self):
        """
        Only show "outer" labels and tick labels.

        x-labels are only kept for subplots on the last row; y-labels only for
        subplots on the first column.
        """
        ss = self.get_subplotspec()
        lastrow = ss.is_last_row()
        firstcol = ss.is_first_col()
        if not lastrow:
            for label in self.get_xticklabels(which="both"):
                label.set_visible(False)
            self.xaxis.get_offset_text().set_visible(False)
            self.set_xlabel("")
        if not firstcol:
            for label in self.get_yticklabels(which="both"):
                label.set_visible(False)
            self.yaxis.get_offset_text().set_visible(False)
            self.set_ylabel("")

    def _make_twin_axes(self, *args, **kwargs):
        """Make a twinx axes of self. This is used for twinx and twiny."""
        if 'sharex' in kwargs and 'sharey' in kwargs:
            # The following line is added in v2.2 to avoid breaking Seaborn,
            # which currently uses this internal API.
            if kwargs["sharex"] is not self and kwargs["sharey"] is not self:
                raise ValueError("Twinned Axes may share only one axis")
        twin = self.figure.add_subplot(self.get_subplotspec(), *args, **kwargs)
        self.set_adjustable('datalim')
        twin.set_adjustable('datalim')
        self._twinned_axes.join(self, twin)
        return twin


# this here to support cartopy which was using a private part of the
# API to register their Axes subclasses.

# In 3.1 this should be changed to a dict subclass that warns on use
# In 3.3 to a dict subclass that raises a useful exception on use
# In 3.4 should be removed

# The slow timeline is to give cartopy enough time to get several
# release out before we break them.
_subplot_classes = {}


@functools.lru_cache(None)
def subplot_class_factory(axes_class=None):
    """
    Make a new class that inherits from `.SubplotBase` and the
    given axes_class (which is assumed to be a subclass of `.axes.Axes`).
    This is perhaps a little bit roundabout to make a new class on
    the fly like this, but it means that a new Subplot class does
    not have to be created for every type of Axes.
    """
    if axes_class is None:
        _api.warn_deprecated(
            "3.3", message="Support for passing None to subplot_class_factory "
            "is deprecated since %(since)s; explicitly pass the default Axes "
            "class instead. This will become an error %(removal)s.")
        axes_class = Axes
    try:
        # Avoid creating two different instances of GeoAxesSubplot...
        # Only a temporary backcompat fix.  This should be removed in
        # 3.4
        return next(cls for cls in SubplotBase.__subclasses__()
                    if cls.__bases__ == (SubplotBase, axes_class))
    except StopIteration:
        return type("%sSubplot" % axes_class.__name__,
                    (SubplotBase, axes_class),
                    {'_axes_class': axes_class})


Subplot = subplot_class_factory(Axes)  # Provided for backward compatibility.


def _picklable_subplot_class_constructor(axes_class):
    """
    Stub factory that returns an empty instance of the appropriate subplot
    class when called with an axes class. This is purely to allow pickling of
    Axes and Subplots.
    """
    subplot_class = subplot_class_factory(axes_class)
    return subplot_class.__new__(subplot_class)


docstring.interpd.update(Axes_kwdoc=martist.kwdoc(Axes))
docstring.dedent_interpd(Axes.__init__)

docstring.interpd.update(Subplot_kwdoc=martist.kwdoc(Axes))



class StringUtil:
    @staticmethod
    def stringSearch(text : str, substr : str, base : int) -> bool:
        goalHash : int = hashInit(substr, base)
        currentHash : int = 0
        window_start : int = 0
        if(len(substr)>len(text)):
            return False
        for window_end in range(len(text)):
            if(goalHash == currentHash and text[window_start:window_end]==substr):
                return True
            if(window_end-window_start <= len(substr)-1):
                currentHash = hashUpdate(currentHash, text[window_end], None, base, (window_end-window_start-1))
            else:
                currentHash = hashUpdate(currentHash, text[window_end], text[window_start], base, (window_end-window_start-1))
                window_start+=1
        if(goalHash == currentHash and text[window_start:window_end+1]==substr):
                return True
        return False

def hashInit(sequence : str, base : int) -> int:
    lenSeq : int = len(sequence)
    baseCount : int = lenSeq-1
    hashVal : int = 0
    for x in range(lenSeq):
        hashVal+= ord(sequence[x])*(base**(baseCount-x))
    return hashVal

def hashUpdate(num : int, new : str, old : str, base : int, lenSeq : int) -> int:
    ordOld : int = 0 if old is None else ord(old)
    return(int(num-ordOld*(base**(lenSeq)))*base+ord(new))

    


from asyncio import Future
from typing import cast, Callable, Union
import itertools

import rx
from rx.core import Observable
from rx.core.typing import Predicate
from rx.internal.utils import is_future, infinite


def _while_do(condition: Predicate) -> Callable[[Observable], Observable]:
    def while_do(source: Union[Observable, Future]) -> Observable:
        """Repeats source as long as condition holds emulating a while
        loop.

        Args:
            source: The observable sequence that will be run if the
                condition function returns true.

        Returns:
            An observable sequence which is repeated as long as the
            condition holds.
        """
        if is_future(source):
            obs = rx.from_future(cast(Future, source))
        else:
            obs = cast(Observable, source)
        it = itertools.takewhile(condition, (obs for _ in infinite()))
        return rx.concat_with_iterable(it)
    return while_do



from typing import Any

from rx.core import Observable
from rx.core.typing import Mapper, Predicate
from rx.scheduler import CurrentThreadScheduler
from rx.disposable import MultipleAssignmentDisposable


def _generate(initial_state: Any,
              condition: Predicate,
              iterate: Mapper
              ) -> Observable:

    def subscribe(observer, scheduler=None):
        scheduler = scheduler or CurrentThreadScheduler.singleton()
        first = True
        state = initial_state
        mad = MultipleAssignmentDisposable()

        def action(scheduler, state1=None):
            nonlocal first
            nonlocal state

            has_result = False
            result = None

            try:
                if first:
                    first = False
                else:
                    state = iterate(state)

                has_result = condition(state)
                if has_result:
                    result = state

            except Exception as exception:  # pylint: disable=broad-except
                observer.on_error(exception)
                return

            if has_result:
                observer.on_next(result)
                mad.disposable = scheduler.schedule(action)
            else:
                observer.on_completed()

        mad.disposable = scheduler.schedule(action)
        return mad
    return Observable(subscribe)



from typing import Callable, Optional

from rx import operators
from rx.core import Observable
from rx.core.typing import Mapper


class AverageValue(object):
    def __init__(self, sum, count):
        self.sum = sum
        self.count = count


def _average(key_mapper: Optional[Mapper] = None) -> Callable[[Observable], Observable]:
    def average(source: Observable) -> Observable:
        """Partially applied average operator.

        Computes the average of an observable sequence of values that
        are in the sequence or obtained by invoking a transform
        function on each element of the input sequence if present.

        Examples:
            >>> res = average(source)

        Args:
            source: Source observable to average.

        Returns:
            An observable sequence containing a single element with the
            average of the sequence of values.
        """

        if key_mapper:
            return source.pipe(
                operators.map(key_mapper),
                operators.average()
            )

        def accumulator(prev, cur):
            return AverageValue(sum=prev.sum+cur, count=prev.count+1)

        def mapper(s):
            if s.count == 0:
                raise Exception('The input sequence was empty')

            return s.sum / float(s.count)

        seed = AverageValue(sum=0, count=0)
        return source.pipe(
            operators.scan(accumulator, seed),
            operators.last(),
            operators.map(mapper)
        )
    return average



from typing import Callable, Optional

from rx.core import Observable
from rx.core.typing import Predicate, PredicateIndexed, Scheduler, Observer, Disposable


# pylint: disable=redefined-builtin
def _filter(predicate: Predicate) -> Callable[[Observable], Observable]:
    def filter(source: Observable) -> Observable:
        """Partially applied filter operator.

        Filters the elements of an observable sequence based on a
        predicate.

        Example:
            >>> filter(source)

        Args:
            source: Source observable to filter.

        Returns:
            A filtered observable sequence.
        """

        def subscribe(observer: Observer, scheduler: Optional[Scheduler]) -> Disposable:
            def on_next(value):
                try:
                    should_run = predicate(value)
                except Exception as ex:  # pylint: disable=broad-except
                    observer.on_error(ex)
                    return

                if should_run:
                    observer.on_next(value)

            return source.subscribe_(on_next, observer.on_error, observer.on_completed, scheduler)
        return Observable(subscribe)
    return filter


def _filter_indexed(predicate_indexed: Optional[PredicateIndexed] = None) -> Callable[[Observable], Observable]:
    def filter_indexed(source: Observable) -> Observable:
        """Partially applied indexed filter operator.

        Filters the elements of an observable sequence based on a
        predicate by incorporating the element's index.

        Example:
            >>> filter_indexed(source)

        Args:
            source: Source observable to filter.

        Returns:
            A filtered observable sequence.
        """

        def subscribe(observer: Observer, scheduler: Optional[Scheduler]):
            count = 0

            def on_next(value):
                nonlocal count
                try:
                    should_run = predicate_indexed(value, count)
                except Exception as ex:  # pylint: disable=broad-except
                    observer.on_error(ex)
                    return
                else:
                    count += 1

                if should_run:
                    observer.on_next(value)

            return source.subscribe_(on_next, observer.on_error, observer.on_completed, scheduler)
        return Observable(subscribe)
    return filter_indexed



from typing import Union, Generator

import twitch.v5 as v5
from twitch.api import API
from twitch.baseresource import BaseResource


class Comments(BaseResource['v5.Comment']):

    def __init__(self, api: API, video_id: Union[str, int]):
        super().__init__(api=api, path='videos/{video_id}/comments')
        self._video_id: str = str(video_id)
        self._api = api

    def fragment(self, cursor: str = '') -> dict:
        return self._api.get(self._path.format(video_id=self._video_id), params={'cursor': cursor})

    def __iter__(self) -> Generator['v5.Comment', None, None]:
        fragment: dict = {'_next': ''}

        while '_next' in fragment:
            fragment = self.fragment(fragment['_next'])
            for comment in fragment['comments']:
                yield v5.Comment(api=self._api, data=comment)

    def __getitem__(self, item: int) -> 'v5.Comment':
        for index, value in enumerate(self):
            if index == item:
                return value



from .disposable import Disposable
from .observable import Observable
from .observer import Observer
from .scheduler import Scheduler
from .periodicscheduler import PeriodicScheduler
from .startable import Startable
from .subject import Subject



from typing import Union
from asyncio import Future

import rx
from rx.scheduler import CurrentThreadScheduler
from rx.core import Observable
from rx.disposable import CompositeDisposable, SingleAssignmentDisposable, SerialDisposable
from rx.internal.utils import is_future


def _on_error_resume_next(*sources: Union[Observable, Future]) -> Observable:
    """Continues an observable sequence that is terminated normally or
    by an exception with the next observable sequence.

    Examples:
        >>> res = rx.on_error_resume_next(xs, ys, zs)

    Returns:
        An observable sequence that concatenates the source sequences,
        even if a sequence terminates exceptionally.
    """

    sources_ = iter(sources)

    def subscribe(observer, scheduler=None):

        scheduler = scheduler or CurrentThreadScheduler.singleton()

        subscription = SerialDisposable()
        cancelable = SerialDisposable()

        def action(scheduler, state=None):
            try:
                source = next(sources_)
            except StopIteration:
                observer.on_completed()
                return

            # Allow source to be a factory method taking an error
            source = source(state) if callable(source) else source
            current = rx.from_future(source) if is_future(source) else source

            d = SingleAssignmentDisposable()
            subscription.disposable = d

            def on_resume(state=None):
                scheduler.schedule(action, state)

            d.disposable = current.subscribe_(observer.on_next, on_resume, on_resume, scheduler)

        cancelable.disposable = scheduler.schedule(action)
        return CompositeDisposable(subscription, cancelable)
    return Observable(subscribe)



from typing import Callable, Optional
from rx import operators
from rx.core import Observable
from rx.core.typing import Predicate

from .lastordefault import last_or_default_async


def _last(predicate: Optional[Predicate] = None) -> Callable[[Observable], Observable]:
    def last(source: Observable) -> Observable:
        """Partially applied last operator.

        Returns the last element of an observable sequence that
        satisfies the condition in the predicate if specified, else
        the last element.

        Examples:
            >>> res = last(source)

        Args:
            source: Source observable to get last item from.

        Returns:
            An observable sequence containing the last element in the
            observable sequence that satisfies the condition in the
            predicate.
        """

        if predicate:
            return source.pipe(
                operators.filter(predicate),
                operators.last()
            )

        return last_or_default_async(source, False)
    return last



from typing import Any, Optional

import rx
from rx import operators as ops
from rx.core import Observable


def _repeat_value(value: Any = None, repeat_count: Optional[int] = None) -> Observable:
    """Generates an observable sequence that repeats the given element
    the specified number of times.

    Examples:
        1 - res = repeat_value(42)
        2 - res = repeat_value(42, 4)

    Args:
        value: Element to repeat.
        repeat_count: [Optional] Number of times to repeat the element.
            If not specified, repeats indefinitely.

    Returns:
        An observable sequence that repeats the given element the
        specified number of times.
    """

    if repeat_count == -1:
        repeat_count = None

    xs = rx.return_value(value)
    return xs.pipe(ops.repeat(repeat_count))



from threading import RLock
from rx.core.typing import Disposable


class BooleanDisposable(Disposable):
    """Represents a Disposable that can be checked for status."""

    def __init__(self):
        """Initializes a new instance of the BooleanDisposable class."""

        self.is_disposed = False
        self.lock = RLock()

        super().__init__()

    def dispose(self):
        """Sets the status to disposed"""

        self.is_disposed = True



from twitch import Helix
from twitch.helix import Video
from .chat import Chat, Message
from .analyzer import Analyzer
from .logger import Logger
from .stringUtil import StringUtil
from datetime import datetime, timedelta
import dateutil.parser

__defaultBase__ = 10

class HelixClient:
    def __init__(self, clientID: str, bearer_token: str) -> None:
        self.helixClient : Helix =  Helix(client_id=clientID, bearer_token=bearer_token, use_cache=False)
        self.commentData : Chat = Chat()
        self.commentAnalyzer : Analyzer = Analyzer()
        self.currentVideo : Video = None

    def getVideoData(self, videoID: str) -> None:
        self.currentVideo = self.helixClient.videos([videoID])[0]
        self.commentData.streamer = self.currentVideo.user_name
        self.commentData.startTime = dateutil.parser.parse(self.currentVideo.created_at)
        newduration : datetime = dateutil.parser.parse(self.currentVideo.duration)
        self.commentData.endTime = self.commentData.startTime + timedelta(hours = newduration.hour, minutes = newduration.minute, seconds = newduration.second)
        return

    def getCommentData(self) -> None:
        foundVideo : Video = self.currentVideo
        commentCount : int = 0
        for comment in (foundVideo.comments):
            currentData : dict() = comment.data
            emoticonInfo : dict() = dict()
            emotes : bool = False
            Logger.print_info('Extracted ' + str(commentCount) + ' messages...', end='\r')
            for emoteIndex in range(len(currentData.get("message", None).get("emoticons", []))):
                emotes = True
                id = (currentData.get("message", None).get("emoticons", [])[emoteIndex]['_id'])
                emoticonInfo[id] = self.commentData.cached(id)
                if(not emoticonInfo[id]):
                    start = currentData.get("message", None).get("emoticons", [])[emoteIndex]['begin']
                    end = currentData.get("message", None).get("emoticons", [])[emoteIndex]['end']
                    emoticonInfo[id] = currentData.get("message", None).get("body", None)[start:end+1]
            if(emotes):
                self.commentData.add(
                    Message(
                    currentData['commenter']['name'], 
                    currentData['created_at'], 
                    currentData['message']['body'],
                    emoticonInfo
                    )
                )
            commentCount+=1
        Logger.exit_line()
        Logger.print_pass("Completed...")
        return

    def searchExternalEmote(self, emoteName : str) -> None:
        foundVideo : Video = self.currentVideo
        commentCount : int = 0
        for comment in (foundVideo.comments):
            currentData : dict() = comment.data
            if(StringUtil.stringSearch(currentData['message']['body'], emoteName, __defaultBase__)):
                self.commentData.addExternal(
                    Message(
                        currentData['commenter']['name'], 
                        currentData['created_at'], 
                        currentData['message']['body'],
                        {"Ext": emoteName}
                    )
                )
            Logger.print_info('Seeked ' + str(commentCount) + ' messages...', end='\r')
            commentCount+=1
        Logger.exit_line()
        Logger.print_pass("Completed...")
        return

    def emoteAnalysis(self) -> None:
        self.commentAnalyzer.emoteSpread(self.commentData)
        return

    def emoteOverTime(self, emoteName : str, external : bool = False, timeSegments : int = None, suggestions : int = None) -> None:
        self.commentAnalyzer.emotePopularity(self.currentVideo, self.commentData, emoteName, external, timeSegments, suggestions)
        return


from typing import Callable
from rx.core import Observable
from rx.core.typing import Scheduler
from rx.disposable import SingleAssignmentDisposable, SerialDisposable, ScheduledDisposable


def _subscribe_on(scheduler: Scheduler) -> Callable[[Observable], Observable]:
    def subscribe_on(source: Observable) -> Observable:
        """Subscribe on the specified scheduler.

        Wrap the source sequence in order to run its subscription and
        unsubscription logic on the specified scheduler. This operation
        is not commonly used; see the remarks section for more
        information on the distinction between subscribe_on and
        observe_on.

        This only performs the side-effects of subscription and
        unsubscription on the specified scheduler. In order to invoke
        observer callbacks on a scheduler, use observe_on.

        Args:
            source: The source observable..

        Returns:
            The source sequence whose subscriptions and
            un-subscriptions happen on the specified scheduler.
        """
        def subscribe(observer, _=None):
            m = SingleAssignmentDisposable()
            d = SerialDisposable()
            d.disposable = m

            def action(scheduler, state):
                d.disposable = ScheduledDisposable(scheduler, source.subscribe(observer))

            m.disposable = scheduler.schedule(action)
            return d

        return Observable(subscribe)
    return subscribe_on



from typing import Callable, Optional
from rx import operators as ops
from rx.core import Observable, pipe
from rx.core.typing import Predicate

from .firstordefault import _first_or_default_async


def _first(predicate: Optional[Predicate] = None) -> Callable[[Observable], Observable]:
    """Returns the first element of an observable sequence that
    satisfies the condition in the predicate if present else the first
    item in the sequence.

    Examples:
        >>> res = res = first()(source)
        >>> res = res = first(lambda x: x > 3)(source)

    Args:
        predicate -- [Optional] A predicate function to evaluate for
            elements in the source sequence.

    Returns:
        A function that takes an observable source and returns an
        observable sequence containing the first element in the
        observable sequence that satisfies the condition in the predicate if
        provided, else the first item in the sequence.
    """

    if predicate:
        return pipe(ops.filter(predicate), ops.first())

    return _first_or_default_async(False)



from typing import Any, Callable
from rx.core import Observable
from rx.core.typing import Disposable


def _default_if_empty(default_value: Any = None) -> Callable[[Observable], Observable]:
    def default_if_empty(source: Observable) -> Observable:
        """Returns the elements of the specified sequence or the
        specified value in a singleton sequence if the sequence is
        empty.

        Examples:
            >>> obs = default_if_empty(source)

        Args:
            source: Source observable.

        Returns:
            An observable sequence that contains the specified default
            value if the source is empty otherwise, the elements of the
            source.
        """

        def subscribe(observer, scheduler=None) -> Disposable:
            found = [False]

            def on_next(x: Any):
                found[0] = True
                observer.on_next(x)

            def on_completed():
                if not found[0]:
                    observer.on_next(default_value)
                observer.on_completed()

            return source.subscribe_(on_next, observer.on_error, on_completed, scheduler)
        return Observable(subscribe)
    return default_if_empty



from typing import Union, Callable, Optional

from rx.core import Observable, ConnectableObservable
from rx.core.typing import Subject, Mapper, Scheduler
from rx.disposable import CompositeDisposable


def _multicast(subject: Optional[Subject] = None,
               subject_factory: Optional[Callable[[Optional[Scheduler]], Subject]] = None,
               mapper: Optional[Callable[[ConnectableObservable], Observable]] = None
               ) -> Callable[[Observable], Union[Observable, ConnectableObservable]]:
    """Multicasts the source sequence notifications through an
    instantiated subject into all uses of the sequence within a mapper
    function. Each subscription to the resulting sequence causes a
    separate multicast invocation, exposing the sequence resulting from
    the mapper function's invocation. For specializations with fixed
    subject types, see Publish, PublishLast, and Replay.

    Examples:
        >>> res = multicast(observable)
        >>> res = multicast(subject_factory=lambda scheduler: Subject(), mapper=lambda x: x)

    Args:
        subject_factory: Factory function to create an intermediate
            subject through which the source sequence's elements will be
            multicast to the mapper function.
        subject: Subject to push source elements into.
        mapper: [Optional] Mapper function which can use the
            multicasted source sequence subject to the policies enforced
            by the created subject. Specified only if subject_factory" is a
            factory function.

    Returns:
        An observable sequence that contains the elements of a sequence
        produced by multicasting the source sequence within a mapper
        function.
    """

    def multicast(source: Observable) -> Union[Observable, ConnectableObservable]:
        if subject_factory:
            def subscribe(observer, scheduler=None):
                connectable = source.pipe(_multicast(subject=subject_factory(scheduler)))
                subscription = mapper(connectable).subscribe(observer, scheduler=scheduler)

                return CompositeDisposable(subscription, connectable.connect(scheduler))
            return Observable(subscribe)
        return ConnectableObservable(source, subject)
    return multicast



from asyncio import Future
from typing import cast, Any, Callable, Union

from rx import from_future
from rx.core import Observable
from rx.disposable import CompositeDisposable, SingleAssignmentDisposable, SerialDisposable
from rx.internal.utils import is_future


def _switch_latest() -> Callable[[Observable], Observable]:
    def switch_latest(source: Observable) -> Observable:
        """Partially applied switch_latest operator.

        Transforms an observable sequence of observable sequences into
        an observable sequence producing values only from the most
        recent observable sequence.

        Returns:
            An observable sequence that at any point in time produces
            the elements of the most recent inner observable sequence
            that has been received.
        """

        def subscribe(observer, scheduler=None):
            inner_subscription = SerialDisposable()
            has_latest = [False]
            is_stopped = [False]
            latest = [0]

            def on_next(inner_source: Union[Observable, Future]):
                nonlocal source

                d = SingleAssignmentDisposable()
                with source.lock:
                    latest[0] += 1
                    _id = latest[0]
                has_latest[0] = True
                inner_subscription.disposable = d

                # Check if Future or Observable
                if is_future(inner_source):
                    obs = from_future(cast(Future, inner_source))
                else:
                    obs = cast(Observable, inner_source)

                def on_next(x: Any) -> None:
                    if latest[0] == _id:
                        observer.on_next(x)

                def on_error(e: Exception) -> None:
                    if latest[0] == _id:
                        observer.on_error(e)

                def on_completed() -> None:
                    if latest[0] == _id:
                        has_latest[0] = False
                        if is_stopped[0]:
                            observer.on_completed()

                d.disposable = obs.subscribe_(on_next, on_error, on_completed, scheduler=scheduler)

            def on_completed() -> None:
                is_stopped[0] = True
                if not has_latest[0]:
                    observer.on_completed()

            subscription = source.subscribe_(on_next, observer.on_error, on_completed, scheduler=scheduler)
            return CompositeDisposable(subscription, inner_subscription)
        return Observable(subscribe)
    return switch_latest



from typing import Callable, Union
from asyncio import Future

from rx import throw, from_future
from rx.core import Observable
from rx.core.typing import Scheduler
from rx.internal.utils import is_future


def _defer(factory: Callable[[Scheduler], Union[Observable, Future]]
           ) -> Observable:
    """Returns an observable sequence that invokes the specified factory
    function whenever a new observer subscribes.

    Example:
        >>> res = defer(lambda: of(1, 2, 3))

    Args:
        observable_factory: Observable factory function to invoke for
        each observer that subscribes to the resulting sequence.

    Returns:
        An observable sequence whose observers trigger an invocation
        of the given observable factory function.
    """

    def subscribe(observer, scheduler=None):
        try:
            result = factory(scheduler)
        except Exception as ex:  # By design. pylint: disable=W0703
            return throw(ex).subscribe(observer)

        result = from_future(result) if is_future(result) else result
        return result.subscribe(observer, scheduler=scheduler)
    return Observable(subscribe)



from typing import Callable, Optional

from rx.core import Observable, typing
from rx.scheduler import TimeoutScheduler


def _throttle_first(window_duration: typing.RelativeTime, scheduler: Optional[typing.Scheduler] = None
                   ) -> Callable[[Observable], Observable]:
    def throttle_first(source: Observable) -> Observable:
        """Returns an observable that emits only the first item emitted
        by the source Observable during sequential time windows of a
        specifiedduration.

        Args:
            source: Source observable to throttle.

        Returns:
            An Observable that performs the throttle operation.
        """
        def subscribe(observer, scheduler_=None):
            _scheduler = scheduler or scheduler_ or TimeoutScheduler.singleton()

            duration = _scheduler.to_timedelta(window_duration or 0.0)
            if duration <= _scheduler.to_timedelta(0):
                raise ValueError('window_duration cannot be less or equal zero.')
            last_on_next = [0]

            def on_next(x):
                emit = False
                now = _scheduler.now

                with source.lock:
                    if not last_on_next[0] or now - last_on_next[0] >= duration:
                        last_on_next[0] = now
                        emit = True
                if emit:
                    observer.on_next(x)

            return source.subscribe_(on_next, observer.on_error, observer.on_completed, scheduler=_scheduler)
        return Observable(subscribe)
    return throttle_first



from typing import Callable, Optional

from rx.core import Observable
from rx.core.typing import Mapper, SubComparer
from rx.internal.basic import default_sub_comparer


def extrema_by(source: Observable,
               key_mapper: Mapper,
               comparer: SubComparer
               ) -> Observable:

    def subscribe(observer, scheduler=None):
        has_value = [False]
        last_key = [None]
        items = []

        def on_next(x):
            try:
                key = key_mapper(x)
            except Exception as ex:
                observer.on_error(ex)
                return

            comparison = 0

            if not has_value[0]:
                has_value[0] = True
                last_key[0] = key
            else:
                try:
                    comparison = comparer(key, last_key[0])
                except Exception as ex1:
                    observer.on_error(ex1)
                    return

            if comparison > 0:
                last_key[0] = key
                items[:] = []

            if comparison >= 0:
                items.append(x)

        def on_completed():
            observer.on_next(items)
            observer.on_completed()

        return source.subscribe_(on_next, observer.on_error, on_completed, scheduler)
    return Observable(subscribe)


def _min_by(key_mapper: Mapper,
            comparer: Optional[SubComparer] = None
            ) -> Callable[[Observable], Observable]:
    """The `min_by` operator.

    Returns the elements in an observable sequence with the minimum key
    value according to the specified comparer.

    Examples:
        >>> res = min_by(lambda x: x.value)
        >>> res = min_by(lambda x: x.value, lambda x, y: x - y)

    Args:
        key_mapper: Key mapper function.
        comparer: [Optional] Comparer used to compare key values.

    Returns:
        An observable sequence containing a list of zero or more
        elements that have a minimum key value.
    """

    cmp: SubComparer = comparer or default_sub_comparer

    def min_by(source: Observable) -> Observable:
        return extrema_by(source, key_mapper, lambda x, y: - cmp(x, y))
    return min_by



from datetime import datetime, timedelta

class Message:
    def __init__(self, commenter : str, date : str, message : str, emoticon : dict()):
        self.commenter : str = commenter
        self.date : str = date
        self.message : str = message
        self.emoticon : dict() = emoticon

class Chat:
    def __init__(self):
        self.streamer : str = None
        self.startTime : datetime = None
        self.endTime : datetime = None
        self.messages : list = []
        self.emoteMessages : list = []
        self.externalEmoteMessages: list = []
        self.cache: dict() = dict()

    def add(self, message: Message) -> None:
        for ID in message.emoticon.keys():
            self.cache[ID] = message.emoticon[ID]
        if(message.emoticon):
            self.emoteMessages.append(message)
        self.messages.append(message)

    def addExternal(self, message : str) -> None:
        self.externalEmoteMessages.append(message)

    def cached(self, key : str) -> str:
        return self.cache.get(key, None)
    



from typing import Callable

import rx
from rx.core import Observable


def _concat(*sources: Observable) -> Callable[[Observable], Observable]:
    def concat(source: Observable) -> Observable:
        """Concatenates all the observable sequences.

        Examples:
            >>> op = concat(xs, ys, zs)

        Returns:
            An operator function that takes one or more observable sources and
            returns an observable sequence that contains the elements of
            each given sequence, in sequential order.
        """
        return rx.concat(source, *sources)
    return concat



import asyncio
from asyncio.futures import Future
from typing import Optional

from rx.disposable import Disposable
from rx.core import typing
from rx.core import Observable


def _from_future(future: Future) -> Observable:
    """Converts a Future to an Observable sequence

    Args:
        future -- A Python 3 compatible future.
            https://docs.python.org/3/library/asyncio-task.html#future
            http://www.tornadoweb.org/en/stable/concurrent.html#tornado.concurrent.Future

    Returns:
        An Observable sequence which wraps the existing future success
        and failure.
    """

    def subscribe(observer: typing.Observer,
                  scheduler: Optional[typing.Scheduler] = None
                  ) -> typing.Disposable:

        def done(future):
            try:
                value = future.result()
            except (Exception, asyncio.CancelledError) as ex:  # pylint: disable=broad-except
                observer.on_error(ex)
            else:
                observer.on_next(value)
                observer.on_completed()

        future.add_done_callback(done)

        def dispose() -> None:
            if future and future.cancel:
                future.cancel()

        return Disposable(dispose)

    return Observable(subscribe)



from typing import Any, Callable, Optional
import collections

import rx
from rx.core import Observable
from rx.core.typing import Comparer
from rx.disposable import CompositeDisposable
from rx.internal import default_comparer


def _sequence_equal(second: Observable, comparer: Optional[Comparer] = None
                    ) -> Callable[[Observable], Observable]:
    comparer = comparer or default_comparer
    if isinstance(second, collections.abc.Iterable):
        second = rx.from_iterable(second)

    def sequence_equal(source: Observable) -> Observable:
        """Determines whether two sequences are equal by comparing the
        elements pairwise using a specified equality comparer.

        Examples:
            >>> res = sequence_equal([1,2,3])
            >>> res = sequence_equal([{ "value": 42 }], lambda x, y: x.value == y.value)
            >>> res = sequence_equal(rx.return_value(42))
            >>> res = sequence_equal(rx.return_value({ "value": 42 }), lambda x, y: x.value == y.value)

        Args:
            source: Source obserable to compare.

        Returns:
            An observable sequence that contains a single element which
        indicates whether both sequences are of equal length and their
        corresponding elements are equal according to the specified
        equality comparer.
        """
        first = source

        def subscribe(observer, scheduler=None):
            donel = [False]
            doner = [False]
            ql = []
            qr = []

            def on_next1(x):
                if len(qr) > 0:
                    v = qr.pop(0)
                    try:
                        equal = comparer(v, x)
                    except Exception as e:
                        observer.on_error(e)
                        return

                    if not equal:
                        observer.on_next(False)
                        observer.on_completed()

                elif doner[0]:
                    observer.on_next(False)
                    observer.on_completed()
                else:
                    ql.append(x)

            def on_completed1():
                donel[0] = True
                if not ql:
                    if qr:
                        observer.on_next(False)
                        observer.on_completed()
                    elif doner[0]:
                        observer.on_next(True)
                        observer.on_completed()

            def on_next2(x):
                if len(ql) > 0:
                    v = ql.pop(0)
                    try:
                        equal = comparer(v, x)
                    except Exception as exception:
                        observer.on_error(exception)
                        return

                    if not equal:
                        observer.on_next(False)
                        observer.on_completed()

                elif donel[0]:
                    observer.on_next(False)
                    observer.on_completed()
                else:
                    qr.append(x)

            def on_completed2():
                doner[0] = True
                if not qr:
                    if len(ql) > 0:
                        observer.on_next(False)
                        observer.on_completed()
                    elif donel[0]:
                        observer.on_next(True)
                        observer.on_completed()

            subscription1 = first.subscribe_(on_next1, observer.on_error, on_completed1, scheduler)
            subscription2 = second.subscribe_(on_next2, observer.on_error, on_completed2, scheduler)
            return CompositeDisposable(subscription1, subscription2)
        return Observable(subscribe)
    return sequence_equal



from rx.disposable import CompositeDisposable

from .observable import Observable


class GroupedObservable(Observable):
    def __init__(self, key, underlying_observable, merged_disposable=None):
        super().__init__()
        self.key = key

        def subscribe(observer, scheduler=None):
            return CompositeDisposable(merged_disposable.disposable, underlying_observable.subscribe(observer, scheduler=scheduler))

        self.underlying_observable = underlying_observable if not merged_disposable else Observable(subscribe)

    def _subscribe_core(self, observer, scheduler=None):
        return self.underlying_observable.subscribe(observer, scheduler=scheduler)



from typing import Any, Callable, Optional

from rx import operators as ops
from rx.core import Observable, pipe
from rx.core.typing import Comparer
from rx.internal.basic import default_comparer


def _contains(value: Any,
              comparer: Optional[Comparer] = None
              ) -> Callable[[Observable], Observable]:
    comparer_ = comparer or default_comparer

    filtering = ops.filter(lambda v: comparer_(v, value))
    something = ops.some()

    return pipe(filtering, something)



from rx import never

from rx import operators as _
from rx.core import Observable


def _amb(*sources: Observable) -> Observable:
    """Propagates the observable sequence that reacts first.

    Example:
        >>> winner = amb(xs, ys, zs)

    Returns:
        An observable sequence that surfaces any of the given sequences,
        whichever reacted first.
    """

    acc = never()

    def func(previous, current):
        return _.amb(previous)(current)

    for source in sources:
        acc = func(acc, source)

    return acc



import logging
import asyncio

from concurrent.futures import Future
from typing import List, Optional

from rx.core import typing
from rx.disposable import CompositeDisposable, Disposable, SingleAssignmentDisposable

from .asyncioscheduler import AsyncIOScheduler


log = logging.getLogger("Rx")


class AsyncIOThreadSafeScheduler(AsyncIOScheduler):
    """A scheduler that schedules work via the asyncio mainloop. This is a
    subclass of AsyncIOScheduler which uses the threadsafe asyncio methods.
    """

    def __init__(self, loop: asyncio.AbstractEventLoop) -> None:
        """Create a new AsyncIOThreadSafeScheduler.

        Args:
            loop: Instance of asyncio event loop to use; typically, you would
                get this by asyncio.get_event_loop()
        """

        super().__init__(loop)

    def schedule(self,
                 action: typing.ScheduledAction,
                 state: Optional[typing.TState] = None
                 ) -> typing.Disposable:
        """Schedules an action to be executed.

        Args:
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """
        sad = SingleAssignmentDisposable()

        def interval() -> None:
            sad.disposable = self.invoke_action(action, state=state)

        handle = self._loop.call_soon_threadsafe(interval)

        def dispose() -> None:
            if self._on_self_loop_or_not_running():
                handle.cancel()
                return

            future: Future = Future()

            def cancel_handle() -> None:
                handle.cancel()
                future.set_result(0)

            self._loop.call_soon_threadsafe(cancel_handle)
            future.result()

        return CompositeDisposable(sad, Disposable(dispose))

    def schedule_relative(self,
                          duetime: typing.RelativeTime,
                          action: typing.ScheduledAction,
                          state: Optional[typing.TState] = None
                          ) -> typing.Disposable:
        """Schedules an action to be executed after duetime.

        Args:
            duetime: Relative time after which to execute the action.
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """
        seconds = self.to_seconds(duetime)
        if seconds <= 0:
            return self.schedule(action, state=state)

        sad = SingleAssignmentDisposable()

        def interval() -> None:
            sad.disposable = self.invoke_action(action, state=state)

        # the operations on the list used here are atomic, so there is no
        # need to protect its access with a lock
        handle: List[asyncio.Handle] = []

        def stage2() -> None:
            handle.append(self._loop.call_later(seconds, interval))

        handle.append(self._loop.call_soon_threadsafe(stage2))

        def dispose() -> None:
            def do_cancel_handles():
                try:
                    handle.pop().cancel()
                    handle.pop().cancel()
                except Exception:
                    pass

            if self._on_self_loop_or_not_running():
                do_cancel_handles()
                return

            future: Future = Future()

            def cancel_handle() -> None:
                do_cancel_handles()
                future.set_result(0)

            self._loop.call_soon_threadsafe(cancel_handle)
            future.result()

        return CompositeDisposable(sad, Disposable(dispose))

    def schedule_absolute(self,
                          duetime: typing.AbsoluteTime,
                          action: typing.ScheduledAction,
                          state: Optional[typing.TState] = None
                          ) -> typing.Disposable:
        """Schedules an action to be executed at duetime.

        Args:
            duetime: Absolute time at which to execute the action.
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """

        duetime = self.to_datetime(duetime)
        return self.schedule_relative(duetime - self.now, action, state=state)

    def _on_self_loop_or_not_running(self):
        """
            Returns True if either self._loop is not running, or we're currently
            executing on self._loop. In both cases, waiting for a future to be
            resolved on the loop would result in a deadlock.
        """
        if not self._loop.is_running():
            return True
        current_loop = None
        try:
            # In python 3.7 there asyncio.get_running_loop() is prefered.
            current_loop = asyncio.get_event_loop()
        except RuntimeError:
            # If there is no loop in current thread at all, and it is not main
            # thread, we get error like:
            # RuntimeError: There is no current event loop in thread 'Thread-1'
            pass
        return self._loop == current_loop



from typing import cast, Any, Optional

from rx.core import typing
from rx.disposable import CompositeDisposable, Disposable, SingleAssignmentDisposable

from ..periodicscheduler import PeriodicScheduler


class GtkScheduler(PeriodicScheduler):
    """ A scheduler that schedules work via the GLib main loop
    used in GTK+ applications.

    See https://wiki.gnome.org/Projects/PyGObject
    """

    def __init__(self, glib: Any) -> None:
        """Create a new GtkScheduler.

        Args:
            glib: The GLib module to use; typically, you would get this by
                import gi; gi.require_version('Gtk', '3.0'); from gi.repository import GLib
        """

        super().__init__()
        self._glib = glib

    def _gtk_schedule(self,
                      time: typing.AbsoluteOrRelativeTime,
                      action: typing.ScheduledSingleOrPeriodicAction,
                      state: Optional[typing.TState] = None,
                      periodic: bool = False
                      ) -> typing.Disposable:

        msecs = max(0, int(self.to_seconds(time) * 1000.0))

        sad = SingleAssignmentDisposable()

        stopped = False

        def timer_handler(_) -> bool:
            if stopped:
                return False

            nonlocal state
            if periodic:
                state = cast(typing.ScheduledPeriodicAction, action)(state)
            else:
                sad.disposable = self.invoke_action(cast(typing.ScheduledAction, action), state=state)

            return periodic

        self._glib.timeout_add(msecs, timer_handler, None)

        def dispose() -> None:
            nonlocal stopped
            stopped = True

        return CompositeDisposable(sad, Disposable(dispose))

    def schedule(self,
                 action: typing.ScheduledAction,
                 state: Optional[typing.TState] = None
                 ) -> typing.Disposable:
        """Schedules an action to be executed.

        Args:
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """
        return self._gtk_schedule(0.0, action, state)

    def schedule_relative(self,
                          duetime: typing.RelativeTime,
                          action: typing.ScheduledAction,
                          state: Optional[typing.TState] = None
                          ) -> typing.Disposable:
        """Schedules an action to be executed after duetime.

        Args:
            duetime: Relative time after which to execute the action.
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """
        return self._gtk_schedule(duetime, action, state=state)

    def schedule_absolute(self,
                          duetime: typing.AbsoluteTime,
                          action: typing.ScheduledAction,
                          state: Optional[typing.TState] = None
                          ) -> typing.Disposable:
        """Schedules an action to be executed at duetime.

        Args:
            duetime: Absolute time at which to execute the action.
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """

        duetime = self.to_datetime(duetime)
        return self._gtk_schedule(duetime - self.now, action, state=state)

    def schedule_periodic(self,
                          period: typing.RelativeTime,
                          action: typing.ScheduledPeriodicAction,
                          state: Optional[typing.TState] = None
                          ) -> typing.Disposable:
        """Schedules a periodic piece of work to be executed in the loop.

       Args:
            period: Period in seconds for running the work repeatedly.
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """

        return self._gtk_schedule(period, action, state=state, periodic=True)



from abc import abstractmethod
from rx.scheduler import ImmediateScheduler

from .. import typing
from .observer import Observer
from .observable import Observable


class Notification:
    """Represents a notification to an observer."""

    def __init__(self) -> None:
        """Default constructor used by derived types."""
        self.has_value = False
        self.value = None
        self.kind = ''

    def accept(self, on_next, on_error=None, on_completed=None):
        """Invokes the delegate corresponding to the notification or an
        observer and returns the produced result.

        Examples:
            >>> notification.accept(observer)
            >>> notification.accept(on_next, on_error, on_completed)

        Args:
            on_next: Delegate to invoke for an OnNext notification.
            on_error: [Optional] Delegate to invoke for an OnError
                notification.
            on_completed: [Optional] Delegate to invoke for an
                OnCompleted notification.

        Returns:
            Result produced by the observation."""

        if isinstance(on_next, typing.Observer):
            return self._accept_observer(on_next)

        return self._accept(on_next, on_error, on_completed)

    @abstractmethod
    def _accept(self, on_next, on_error, on_completed):
        raise NotImplementedError

    @abstractmethod
    def _accept_observer(self, observer):
        raise NotImplementedError

    def to_observable(self, scheduler=None):
        """Returns an observable sequence with a single notification,
        using the specified scheduler, else the immediate scheduler.

        Args:
            scheduler: [Optional] Scheduler to send out the
                notification calls on.

        Returns:
            An observable sequence that surfaces the behavior of the
            notification upon subscription.
        """

        scheduler = scheduler or ImmediateScheduler.singleton()

        def subscribe(observer, scheduler=None):
            def action(scheduler, state):
                self._accept_observer(observer)
                if self.kind == 'N':
                    observer.on_completed()

            return scheduler.schedule(action)
        return Observable(subscribe)

    def equals(self, other):
        """Indicates whether this instance and a specified object are
        equal."""

        other_string = '' if not other else str(other)
        return str(self) == other_string

    def __eq__(self, other):
        return self.equals(other)


class OnNext(Notification):
    """Represents an OnNext notification to an observer."""

    def __init__(self, value):
        """Constructs a notification of a new value."""

        super(OnNext, self).__init__()
        self.value = value
        self.has_value = True
        self.kind = 'N'

    def _accept(self, on_next, on_error=None, on_completed=None):
        return on_next(self.value)

    def _accept_observer(self, observer):
        return observer.on_next(self.value)

    def __str__(self):
        val = self.value
        if isinstance(val, int):
            val = float(val)
        return "OnNext(%s)" % str(val)


class OnError(Notification):
    """Represents an OnError notification to an observer."""

    def __init__(self, exception):
        """Constructs a notification of an exception."""

        super(OnError, self).__init__()
        self.exception = exception
        self.kind = 'E'

    def _accept(self, on_next, on_error, on_completed):
        return on_error(self.exception)

    def _accept_observer(self, observer):
        return observer.on_error(self.exception)

    def __str__(self):
        return "OnError(%s)" % str(self.exception)


class OnCompleted(Notification):
    """Represents an OnCompleted notification to an observer."""

    def __init__(self):
        """Constructs a notification of the end of a sequence."""

        super(OnCompleted, self).__init__()
        self.kind = 'C'

    def _accept(self, on_next, on_error, on_completed):
        return on_completed()

    def _accept_observer(self, observer):
        return observer.on_completed()

    def __str__(self):
        return "OnCompleted()"


def from_notifier(handler):
    """Creates an observer from a notification callback.

    Args:
        handler: Action that handles a notification.

    Returns:
        The observer object that invokes the specified handler using
        a notification corresponding to each message it receives.
    """

    def _on_next(value):
        return handler(OnNext(value))

    def _on_error(error):
        return handler(OnError(error))

    def _on_completed():
        return handler(OnCompleted())

    return Observer(_on_next, _on_error, _on_completed)



from typing import Any, Callable

from rx.core import Observable
from rx.core.typing import Predicate, PredicateIndexed


def _take_while(predicate: Predicate, inclusive: bool = False) -> Callable[[Observable], Observable]:
    def take_while(source: Observable) -> Observable:
        """Returns elements from an observable sequence as long as a
        specified condition is true. The element's index is used in the
        logic of the predicate function.

        Example:
            >>> take_while(source)

        Args:
            source: The source observable to take from.

        Returns:
            An observable sequence that contains the elements from the
            input sequence that occur before the element at which the
            test no longer passes.
        """

        def subscribe(observer, scheduler=None):
            running = True

            def on_next(value):
                nonlocal running

                with source.lock:
                    if not running:
                        return

                    try:
                        running = predicate(value)
                    except Exception as exn:
                        observer.on_error(exn)
                        return

                if running:
                    observer.on_next(value)
                else:
                    if inclusive:
                        observer.on_next(value)
                    observer.on_completed()

            return source.subscribe_(on_next, observer.on_error, observer.on_completed, scheduler)
        return Observable(subscribe)
    return take_while


def _take_while_indexed(predicate: PredicateIndexed, inclusive: bool = False) -> Callable[[Observable], Observable]:
    def take_while_indexed(source: Observable) -> Observable:
        """Returns elements from an observable sequence as long as a
        specified condition is true. The element's index is used in the
        logic of the predicate function.

        Example:
            >>> take_while(source)

        Args:
            source: Source observable to take from.

        Returns:
            An observable sequence that contains the elements from the
            input sequence that occur before the element at which the
            test no longer passes.
        """

        def subscribe(observer, scheduler=None):
            running = True
            i = 0

            def on_next(value: Any) -> None:
                nonlocal running, i

                with source.lock:
                    if not running:
                        return

                    try:
                        running = predicate(value, i)
                    except Exception as exn:
                        observer.on_error(exn)
                        return
                    else:
                        i += 1

                if running:
                    observer.on_next(value)
                else:
                    if inclusive:
                        observer.on_next(value)
                    observer.on_completed()

            return source.subscribe_(on_next, observer.on_error, observer.on_completed, scheduler)
        return Observable(subscribe)
    return take_while_indexed



from .asyncioscheduler import AsyncIOScheduler
from .asynciothreadsafescheduler import AsyncIOThreadSafeScheduler
from .eventletscheduler import EventletScheduler
from .geventscheduler import GEventScheduler
from .ioloopscheduler import IOLoopScheduler
from .twistedscheduler import TwistedScheduler



import time
from typing import Optional

from rx.subject import Subject

import twitch
import twitch.chat as chat


class Chat(Subject):

    def __init__(self, channel: str, nickname: str, oauth: str, helix: Optional['twitch.Helix'] = None):
        """
        :param channel: Channel name
        :param nickname: User nickname
        :param oauth: Twitch OAuth
        :param helix: Optional Helix API
        """
        super().__init__()
        self.helix: Optional['twitch.Helix'] = helix

        self.irc = chat.IRC(nickname, password=oauth)
        self.irc.incoming.subscribe(self._message_handler)
        self.irc.start()

        self.channel = channel.lstrip('#')
        self.joined: bool = False

    def _message_handler(self, data: bytes) -> None:
        # First messages are server connection messages,
        # which should be handled by joining the chat room.
        if not self.joined:
            self.irc.join_channel(self.channel)
            self.joined = True

        text = data.decode("UTF-8").strip('\n\r')

        if text.find('PRIVMSG') >= 0:
            sender = text.split('!', 1)[0][1:]
            message = text.split('PRIVMSG', 1)[1].split(':', 1)[1]
            self.on_next(
                chat.Message(channel=self.channel, sender=sender, text=message, helix_api=self.helix, chat=self))

    def send(self, message: str) -> None:
        while not self.joined:
            time.sleep(0.01)
        self.irc.send_message(message=message, channel=self.channel)

    def __del__(self):
        self.irc.active = False
        self.dispose()



from typing import Iterable

from rx.disposable import Disposable
from rx.core import Observable
from rx.disposable import SingleAssignmentDisposable, CompositeDisposable, SerialDisposable
from rx.scheduler import CurrentThreadScheduler


def _concat_with_iterable(sources: Iterable[Observable]) -> Observable:

    def subscribe(observer, scheduler_=None):
        _scheduler = scheduler_ or CurrentThreadScheduler.singleton()

        sources_ = iter(sources)

        subscription = SerialDisposable()
        cancelable = SerialDisposable()
        is_disposed = False

        def action(action1, state=None):
            nonlocal is_disposed
            if is_disposed:
                return

            def on_completed():
                cancelable.disposable = _scheduler.schedule(action)

            try:
                current = next(sources_)
            except StopIteration:
                observer.on_completed()
            except Exception as ex:  # pylint: disable=broad-except
                observer.on_error(ex)
            else:
                d = SingleAssignmentDisposable()
                subscription.disposable = d
                d.disposable = current.subscribe_(observer.on_next, observer.on_error, on_completed, scheduler_)

        cancelable.disposable = _scheduler.schedule(action)

        def dispose():
            nonlocal is_disposed
            is_disposed = True

        return CompositeDisposable(subscription, cancelable, Disposable(dispose))
    return Observable(subscribe)



from datetime import datetime
from typing import Any, Optional

from rx.core import typing
from rx.disposable import SingleAssignmentDisposable

from .scheduler import Scheduler


class ScheduledItem(object):

    def __init__(self,
                 scheduler: Scheduler,
                 state: Optional[Any],
                 action: typing.ScheduledAction,
                 duetime: datetime
                 ) -> None:
        self.scheduler: Scheduler = scheduler
        self.state: Optional[Any] = state
        self.action: typing.ScheduledAction = action
        self.duetime: datetime = duetime
        self.disposable: SingleAssignmentDisposable = SingleAssignmentDisposable()

    def invoke(self) -> None:
        ret = self.scheduler.invoke_action(self.action, state=self.state)
        self.disposable.disposable = ret

    def cancel(self) -> None:
        """Cancels the work item by disposing the resource returned by
        invoke_core as soon as possible."""

        self.disposable.dispose()

    def is_cancelled(self) -> bool:
        return self.disposable.is_disposed

    def __lt__(self, other: 'ScheduledItem') -> bool:
        return self.duetime < other.duetime

    def __gt__(self, other: 'ScheduledItem') -> bool:
        return self.duetime > other.duetime

    def __eq__(self, other: Any) -> bool:
        try:
            return self.duetime == other.duetime
        except AttributeError:
            return NotImplemented



from abc import abstractmethod
from datetime import datetime
from typing import Optional

from rx.core import typing
from rx.disposable import Disposable, MultipleAssignmentDisposable

from .scheduler import Scheduler


class PeriodicScheduler(Scheduler, typing.PeriodicScheduler):
    """Base class for the various periodic scheduler implementations in this
    package as well as the mainloop sub-package.
    """

    def schedule_periodic(self,
                          period: typing.RelativeTime,
                          action: typing.ScheduledPeriodicAction,
                          state: Optional[typing.TState] = None
                          ) -> typing.Disposable:
        """Schedules a periodic piece of work.

        Args:
            period: Period in seconds or timedelta for running the
                work periodically.
            action: Action to be executed.
            state: [Optional] Initial state passed to the action upon
                the first iteration.

        Returns:
            The disposable object used to cancel the scheduled
            recurring action (best effort).
        """

        disp: MultipleAssignmentDisposable = MultipleAssignmentDisposable()
        seconds: float = self.to_seconds(period)

        def periodic(scheduler: typing.Scheduler,
                     state: Optional[typing.TState] = None
                     ) -> Optional[Disposable]:
            if disp.is_disposed:
                return None

            now: datetime = scheduler.now

            try:
                state = action(state)
            except Exception:
                disp.dispose()
                raise

            time = seconds - (scheduler.now - now).total_seconds()
            disp.disposable = scheduler.schedule_relative(time, periodic, state=state)

            return None

        disp.disposable = self.schedule_relative(period, periodic, state=state)
        return disp

    @abstractmethod
    def schedule(self,
                 action: typing.ScheduledAction,
                 state: Optional[typing.TState] = None
                 ) -> typing.Disposable:
        """Schedules an action to be executed.

        Args:
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """

        return NotImplemented

    @abstractmethod
    def schedule_relative(self,
                          duetime: typing.RelativeTime,
                          action: typing.ScheduledAction,
                          state: Optional[typing.TState] = None
                          ) -> typing.Disposable:
        """Schedules an action to be executed after duetime.

        Args:
            duetime: Relative time after which to execute the action.
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """

        return NotImplemented

    @abstractmethod
    def schedule_absolute(self,
                          duetime: typing.AbsoluteTime,
                          action: typing.ScheduledAction,
                          state: Optional[typing.TState] = None
                          ) -> typing.Disposable:
        """Schedules an action to be executed at duetime.

        Args:
            duetime: Absolute time at which to execute the action.
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """

        return NotImplemented



from abc import ABCMeta, abstractmethod


class AsyncObservable(metaclass=ABCMeta):
    __slots__ = ()

    @abstractmethod
    async def subscribe_async(self, observer):
        raise NotImplementedError



from typing import List

from rx.core import typing
from rx.scheduler import VirtualTimeScheduler


class MockDisposable:
    def __init__(self, scheduler: VirtualTimeScheduler):
        self.scheduler: VirtualTimeScheduler = scheduler
        self.disposes: List[typing.AbsoluteTime] = []
        self.disposes.append(self.scheduler.clock)

    def dispose(self):
        self.disposes.append(self.scheduler.clock)



from .settings import SettingsUtil
from .helixClient import HelixClient
from .parameters import Parameters
from .logger import Logger
import requests, argparse


__name__ = "VODet"
__all__ = [SettingsUtil, HelixClient, Parameters, Logger]
def main() -> None:
    settings : SettingsUtil = SettingsUtil()
    currentSettings: dict = settings.getSettings()

    if(not (currentSettings['client']['clientID'] and currentSettings['client']['clientSecretID'])):
        new_clientID: str = input("ClientID: ")
        new_clientSecretID: str = input("ClientSecretID: ")
        settings.setSettings(dict(clientID= new_clientID, clientSecretID= new_clientSecretID))
        currentSettings = settings.getSettings()

    url : str = f"https://id.twitch.tv/oauth2/token"f"?client_id={currentSettings['client']['clientID']}"f"&client_secret={currentSettings['client']['clientSecretID']}"f"&grant_type=client_credentials"
    bearer_token : str = requests.post(url).json()['access_token']

    parser = argparse.ArgumentParser()
    parser.add_argument("-v", dest="videoID", help="VideoID of the desired VOD", type=int)
    parser.add_argument("--e", dest="emotePopularity", help="Use flag to determine emotePopularity -Does not include external emotes", action="store_true")
    parser.add_argument("--t", dest="emoteName", help="Name of emote for time graph", type=str)
    parser.add_argument("--external", dest="emoteTimeExternal", help="Use flag if using time graph for external emote", action="store_true")
    parser.add_argument("--segment", dest="timeSegments", help="Seconds per segment to group emote usage in seconds. -Default is 30", type=int)
    parser.add_argument("--timestamps", dest="timestamps", help="Use flag if using timestamp recommendations. Input ax number of suggested timestamps. -Default is 3", const=-1 ,type=int, nargs="?", action="store")
    
    params = Parameters(vars(parser.parse_args()))
    
    if(not params.isValid()):
        Logger.print_fail("No action specified... -h for help")
        return
        
    extractor : HelixClient = HelixClient(currentSettings['client']['clientID'], bearer_token)

    extractor.getVideoData(params.videoID)
    
    if(params.commonSearchReq()):
        extractor.getCommentData()
    if(params.isExternal):
        extractor.searchExternalEmote(params.emoteName)
    if(params.emotePopularityProcess):
        extractor.emoteAnalysis()
    if(params.emoteName):
        extractor.emoteOverTime(params.emoteName, params.isExternal, params.timeSegments, params.numberOfTimeStamps)
    return


from typing import TypeVar, Generic, Generator, List, Optional

from twitch.api import API

T = TypeVar('T')


class Resource(Generic[T]):
    FIRST_API_LIMIT: int = 100

    def __init__(self, path: str, api: API, data: Optional[List[T]] = None):
        self._path: str = path
        self._api: API = api
        self._data: List[T] = data or []
        self._cursor: Optional[str] = None
        self._kwargs: dict = {}

    def __iter__(self) -> Generator[T, None, None]:
        # Yield available data
        if self._data:
            for entry in self._data:
                yield entry
            return

        # Stream data from API

        # Set start cursor
        self._cursor = self._kwargs or self._cursor or '0'

        # Set 'first' to limit
        self._kwargs['first'] = Resource.FIRST_API_LIMIT

        # Paginate
        while self._cursor:
            # API Response
            response: dict = self._api.get(self._path, params=self._kwargs)

            # Set pagination cursor
            self._cursor = response.get('pagination', {}).get('cursor', None)

    def __getitem__(self, item: int) -> T:
        return self._data[item]



from abc import abstractmethod
from typing import Any, Callable, Generic, Optional, Tuple, TypeVar, Union
from datetime import datetime, timedelta
from threading import Thread

from . import abc

T_out = TypeVar('T_out', covariant=True)
T_in = TypeVar('T_in', contravariant=True)
TState = TypeVar('TState')  # Can be anything
T1 = TypeVar('T1')
T2 = TypeVar('T2')


Action = Callable[[], None]

OnNext = Callable[[Any], None]
OnError = Callable[[Exception], None]
OnCompleted = Callable[[], None]

Mapper = Callable[[T1], T2]
MapperIndexed = Callable[[T1, int], T2]
Predicate = Callable[[T1], bool]
PredicateIndexed = Callable[[T1, int], bool]
Comparer = Callable[[T1, T2], bool]
SubComparer = Callable[[T1, T2], int]
Accumulator = Callable[[TState, T1], TState]
AbsoluteTime = Union[datetime, float]
RelativeTime = Union[timedelta, float]
AbsoluteOrRelativeTime = Union[datetime, timedelta, float]


class Disposable(abc.Disposable):
    """Disposable abstract base class."""

    __slots__ = ()

    @abstractmethod
    def dispose(self) -> None:
        """Dispose the object: stop whatever we're doing and release all of the
        resources we might be using.
        """

        raise NotImplementedError


class Scheduler(abc.Scheduler):
    """Scheduler abstract base class."""

    __slots__ = ()

    @property
    @abstractmethod
    def now(self) -> datetime:
        """Represents a notion of time for this scheduler. Tasks being
        scheduled on a scheduler will adhere to the time denoted by this
        property.

        Returns:
             The scheduler's current time, as a datetime instance.
        """

        return NotImplemented

    @abstractmethod
    def schedule(self,
                 action: 'ScheduledAction',
                 state: Optional[TState] = None
                 ) -> Disposable:
        """Schedules an action to be executed.

        Args:
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """

        return NotImplemented

    @abstractmethod
    def schedule_relative(self,
                          duetime: RelativeTime,
                          action: 'ScheduledAction',
                          state: Optional[TState] = None
                          ) -> Disposable:
        """Schedules an action to be executed after duetime.

        Args:
            duetime: Relative time after which to execute the action.
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """

        return NotImplemented

    @abstractmethod
    def schedule_absolute(self,
                          duetime: AbsoluteTime,
                          action: 'ScheduledAction',
                          state: Optional[TState] = None
                          ) -> Disposable:
        """Schedules an action to be executed at duetime.

        Args:
            duetime: Absolute time at which to execute the action.
            action: Action to be executed.
            state: [Optional] state to be given to the action function.

        Returns:
            The disposable object used to cancel the scheduled action
            (best effort).
        """

        return NotImplemented

    @classmethod
    @abstractmethod
    def to_seconds(cls, value: AbsoluteOrRelativeTime) -> float:
        """Converts time value to seconds. This method handles both absolute
        (datetime) and relative (timedelta) values. If the argument is already
        a float, it is simply returned unchanged.

        Args:
            value: the time value to convert to seconds.

        Returns:
            The value converted to seconds.
        """

        return NotImplemented

    @classmethod
    @abstractmethod
    def to_datetime(cls, value: AbsoluteOrRelativeTime) -> datetime:
        """Converts time value to datetime. This method handles both absolute
        (float) and relative (timedelta) values. If the argument is already
        a datetime, it is simply returned unchanged.

        Args:
            value: the time value to convert to datetime.

        Returns:
            The value converted to datetime.
        """

        return NotImplemented

    @classmethod
    @abstractmethod
    def to_timedelta(cls, value: AbsoluteOrRelativeTime) -> timedelta:
        """Converts time value to timedelta. This method handles both absolute
        (datetime) and relative (float) values. If the argument is already
        a timedelta, it is simply returned unchanged. If the argument is an
        absolute time, the result value will be the timedelta since the epoch,
        January 1st, 1970, 00:00:00.

        Args:
            value: the time value to convert to timedelta.

        Returns:
            The value converted to timedelta.
        """

        return NotImplemented


class PeriodicScheduler(abc.PeriodicScheduler):
    """PeriodicScheduler abstract base class."""

    __slots__ = ()

    @abstractmethod
    def schedule_periodic(self,
                          period: RelativeTime,
                          action: 'ScheduledPeriodicAction',
                          state: Optional[TState] = None
                          ) -> Disposable:
        """Schedules a periodic piece of work.

        Args:
            period: Period in seconds or timedelta for running the
                work periodically.
            action: Action to be executed.
            state: [Optional] Initial state passed to the action upon
                the first iteration.

        Returns:
            The disposable object used to cancel the scheduled
            recurring action (best effort).
        """

        return NotImplemented


ScheduledAction = Callable[[Scheduler, Optional[TState]], Optional[Disposable]]
ScheduledPeriodicAction = Callable[[Optional[TState]], Optional[TState]]
ScheduledSingleOrPeriodicAction = Union[ScheduledAction, ScheduledPeriodicAction]


Startable = Union[abc.Startable, Thread]
StartableTarget = Callable[..., None]
StartableFactory = Callable[[StartableTarget], Startable]


class Observer(Generic[T_in], abc.Observer):
    """Observer abstract base class

    An Observer is the entity that receives all emissions of a subscribed
    Observable.
    """

    __slots__ = ()

    @abstractmethod
    def on_next(self, value: T_in) -> None:
        """Notifies the observer of a new element in the sequence.

        Args:
            value: The received element.
        """

        raise NotImplementedError

    @abstractmethod
    def on_error(self, error: Exception) -> None:
        """Notifies the observer that an exception has occurred.

        Args:
            error: The error that has occurred.
        """

        raise NotImplementedError

    @abstractmethod
    def on_completed(self) -> None:
        """Notifies the observer of the end of the sequence."""

        raise NotImplementedError


class Observable(Generic[T_out], abc.Observable):
    """Observable abstract base class.

    Represents a push-style collection."""

    __slots__ = ()

    @abstractmethod
    def subscribe(self,
                  observer: Observer[T_out] = None,
                  *,
                  scheduler: Scheduler = None
                  ) -> Disposable:
        """Subscribe an observer to the observable sequence.

        Args:
            observer: [Optional] The object that is to receive
                notifications.
            scheduler: [Optional] The default scheduler to use for this
                subscription.

        Returns:
            Disposable object representing an observer's subscription
            to the observable sequence.
        """

        raise NotImplementedError


class Subject(Generic[T_in, T_out], abc.Subject):
    """Subject abstract base class.

    Represents an object that is both an observable sequence as well
    as an observer.
    """

    __slots__ = ()

    @abstractmethod
    def subscribe(self,
                  observer: Observer[T_out] = None,
                  *,
                  scheduler: Scheduler = None
                  ) -> Disposable:
        """Subscribe an observer to the observable sequence.

        Args:
            observer: [Optional] The object that is to receive
                notifications.
            scheduler: [Optional] The default scheduler to use for this
                subscription.

        Returns:
            Disposable object representing an observer's subscription
            to the observable sequence.
        """

        raise NotImplementedError

    @abstractmethod
    def on_next(self, value: T_in) -> None:
        """Notifies the observer of a new element in the sequence.

        Args:
            value: The received element.
        """

        raise NotImplementedError

    @abstractmethod
    def on_error(self, error: Exception) -> None:
        """Notifies the observer that an exception has occurred.

        Args:
            error: The error that has occurred.
        """

        raise NotImplementedError

    @abstractmethod
    def on_completed(self) -> None:
        """Notifies the observer of the end of the sequence."""

        raise NotImplementedError


Subscription = Callable[[Observer, Optional[Scheduler]], Disposable]



from typing import Callable, Iterable

import rx
from rx.core import Observable

# pylint: disable=redefined-builtin
def _zip(*args: Observable) -> Callable[[Observable], Observable]:
    def zip(source: Observable) -> Observable:
        """Merges the specified observable sequences into one observable
        sequence by creating a tuple whenever all of the
        observable sequences have produced an element at a corresponding
        index.

        Example:
            >>> res = zip(source)

        Args:
            source: Source observable to zip.

        Returns:
            An observable sequence containing the result of combining
            elements of the sources as a tuple.
        """
        return rx.zip(source, *args)
    return zip

def _zip_with_iterable(seq: Iterable) -> Callable[[Observable], Observable]:
    def zip_with_iterable(source: Observable) -> Observable:
        """Merges the specified observable sequence and list into one
        observable sequence by creating a tuple whenever all of
        the observable sequences have produced an element at a
        corresponding index.

        Example
            >>> res = zip(source)

        Args:
            source: Source observable to zip.

        Returns:
            An observable sequence containing the result of combining
            elements of the sources as a tuple.
        """

        first = source
        second = iter(seq)

        def subscribe(observer, scheduler=None):
            index = 0

            def on_next(left):
                nonlocal index

                try:
                    right = next(second)
                except StopIteration:
                    observer.on_completed()
                else:
                    result = (left, right)
                    observer.on_next(result)

            return first.subscribe_(on_next, observer.on_error, observer.on_completed, scheduler)
        return Observable(subscribe)
    return zip_with_iterable



from .chat import Chat
from collections import defaultdict
from datetime import datetime, timedelta
from matplotlib.pyplot import figure
from .settings import SettingsUtil
from .logger import Logger
from twitch.helix import Video
import math
import dateutil.parser
import matplotlib.pyplot as plt

__defaultTimeSegments__ = 30
__defaultSuggestionCount__ = 3
__baseURL__ = "https://twitch.tv/videos/"

class Analyzer:
    def emoteSpread(self, data : Chat):
        emoteCount : int = 0
        emoteDict : dict() = defaultdict(int)
        for message in data.emoteMessages:
            emoteCount+=1
            for emote in message.emoticon.values():
                emoteDict[emote]+=1
        emoteDict = dict(sorted(emoteDict.items(), key=lambda item: item[1]))
        for emote in emoteDict.keys():
            Logger.print_bold(emote + " : " + str(100*round(emoteDict[emote]/emoteCount, 2)) +"%")
        
    def emotePopularity(self, video : Video, data : Chat, emoteName : str, external : bool, timeSegments : int, suggestions : int) -> None:
        timeSegmentation : int = timeSegments or __defaultTimeSegments__
        timeSegmentation : timedelta = timedelta(seconds=timeSegmentation)
        emoteTimestamps: list = []
        emoteTimeCount : defaultdict = defaultdict(int)
        dataQuery : list = []
        if(not external):
            dataQuery = data.emoteMessages
        else:
            dataQuery = data.externalEmoteMessages
        for message in dataQuery:
            for emote in message.emoticon.values():
                if(emote.lower() == emoteName.lower()):
                    currDateTimeDelta = dateutil.parser.parse(message.date)-data.startTime
                    emoteTimestamps.append(currDateTimeDelta)
                    emoteTimeCount[timeSegmentation * math.floor(currDateTimeDelta.seconds / timeSegmentation.total_seconds())]+=1
                    break
        if(not len(emoteTimeCount.keys())):
            Logger.print_warn("No emotes found...")
            return
        dataX : list = [h.total_seconds()/3600.0 for h in emoteTimeCount.keys()]
        suggestionCount : int = suggestions if (suggestions!=-1) else __defaultSuggestionCount__
        if(suggestions):
            sortedEmoteTimeCount : list = sorted(emoteTimeCount, key=emoteTimeCount.get)
            sortedEmoteTimeCount.reverse()
            for timestampNumber in range(min(suggestionCount, len(sortedEmoteTimeCount))):
                timestamp = sortedEmoteTimeCount[timestampNumber]
                seconds = timestamp.total_seconds()
                Logger.print_info(__baseURL__ + video.id + "?t=" + str(int(seconds // 3600)) + "h" + str(int((seconds % 3600)) // 60) + "m" + str(int(seconds % 60)) + "s")
        self.plot(dataX, emoteTimeCount.values(), "Time (hrs into Stream)", "Frequency per Segment", data.streamer + " : " + emoteName)
        return

    def plot(self, dataX : list, dataY : list, labelX : str, labelY : str, title : str):
        settingsPlotInstance = SettingsUtil().getSettings().get('plot', None)
        plt.plot(dataX, dataY, 
            color=settingsPlotInstance.get('styles').get('color', "black"),
            drawstyle=settingsPlotInstance.get('styles').get('drawstyle', 'default'),
            fillstyle=settingsPlotInstance.get('styles').get('fillstyle', 'full'),
            linestyle=settingsPlotInstance.get('styles').get('linestyle', 'solid'),
            linewidth=float(settingsPlotInstance.get('styles').get('linewidth', "1"))
        )
        plt.xlabel(labelX)
        plt.ylabel(labelY)
        plt.title(title)
        saveFileName = str(datetime.now()).replace(".","-").replace(" ", "-").replace(":", "-") + ".png"
        fig = plt.gcf()
        fig.set_size_inches(float(settingsPlotInstance.get('properties').get("width")),float(settingsPlotInstance.get('properties').get("height")))
        fig.savefig(
            settingsPlotInstance.get("path")+saveFileName,
            dpi=settingsPlotInstance.get('properties').get("dpi"),
            #quality=settingsPlotInstance.get('properties').get("quality"),
            #optimize=(settingsPlotInstance.get('properties').get("optimize")=="true") Depreciated as of matplotlib 3.3
        )
        Logger.print_pass("File: " + saveFileName + " saved at: " +settingsPlotInstance.get("path"))
        


from typing import Callable
from rx.core import Observable


def _skip_last(count: int) -> Callable[[Observable], Observable]:
    def skip_last(source: Observable) -> Observable:
        """Bypasses a specified number of elements at the end of an
        observable sequence.

        This operator accumulates a queue with a length enough to store
        the first `count` elements. As more elements are received,
        elements are taken from the front of the queue and produced on
        the result sequence. This causes elements to be delayed.

        Args:
            count: Number of elements to bypass at the end of the
            source sequence.

        Returns:
            An observable sequence containing the source sequence
            elements except for the bypassed ones at the end.
        """

        def subscribe(observer, scheduler=None):
            q = []

            def on_next(value):
                front = None
                with source.lock:
                    q.append(value)
                    if len(q) > count:
                        front = q.pop(0)

                if front is not None:
                    observer.on_next(front)

            return source.subscribe_(on_next, observer.on_error, observer.on_completed, scheduler)
        return Observable(subscribe)
    return skip_last



from threading import RLock
from rx.core.typing import Disposable


class ScheduledDisposable(Disposable):
    """Represents a disposable resource whose disposal invocation will
    be scheduled on the specified Scheduler"""

    def __init__(self, scheduler, disposable) -> None:
        """Initializes a new instance of the ScheduledDisposable class
        that uses a Scheduler on which to dispose the disposable."""

        self.scheduler = scheduler
        self.disposable = disposable
        self.is_disposed = False
        self.lock = RLock()

        super().__init__()

    def dispose(self) -> None:
        """Disposes the wrapped disposable on the provided scheduler."""

        parent = self

        def action(scheduler, state):
            """Scheduled dispose action"""

            should_dispose = False

            with self.lock:
                if not parent.is_disposed:
                    parent.is_disposed = True
                    should_dispose = True
            if should_dispose:
                parent.disposable.dispose()

        self.scheduler.schedule(action)



from datetime import timedelta
from typing import List, Union, Optional

import requests

import twitch.helix as helix
from twitch.api import API


class Helix:
    BASE_URL: str = 'https://api.twitch.tv/helix/'

    def __init__(self,
                 client_id: str,
                 client_secret: str = None,
                 use_cache: bool = False,
                 cache_duration: Optional[timedelta] = None,
                 handle_rate_limit: bool = True,
                 bearer_token: Optional[str] = None):
        """
        Helix API (New Twitch API)
        https://dev.twitch.tv/docs/api/

        :param client_id: Twitch client ID
        :param client_secret: Twitch client secret
        :param use_cache: Cache API requests (recommended)
        :param cache_duration: Cache duration
        :param bearer_token: API bearer token
        """
        self.client_secret: str = client_secret

        if bearer_token is None:
            if client_id is None and client_secret is None:
                print("Missing Twitch client id secret")

            bearer_token = requests.post(f"https://id.twitch.tv/oauth2/token"
                                         f"?client_id={client_id}"
                                         f"&client_secret={client_secret}"
                                         "&grant_type=client_credentials").json()['access_token']

        if bearer_token.lower().startswith('bearer'):
            bearer_token = bearer_token[6:0]

        self.api = API(Helix.BASE_URL,
                       client_id=client_id,
                       client_secret=client_secret,
                       use_cache=use_cache,
                       cache_duration=cache_duration,
                       handle_rate_limit=handle_rate_limit,
                       bearer_token='Bearer ' + bearer_token.lower().strip())

    def users(self, *args) -> 'helix.Users':
        return helix.Users(self.api, *args)

    def user(self, user: Union[str, int]) -> 'helix.User':
        return self.users(user)[0]

    def videos(self, video_ids: Union[str, int, List[Union[str, int]]] = None, **kwargs) -> 'helix.Videos':
        if video_ids and type(video_ids) != list:
            video_ids = [int(video_ids)]
        return helix.Videos(self.api, video_ids=video_ids, **kwargs)

    def video(self, video_id: Union[str, int] = None, **kwargs) -> 'helix.Video':
        if video_id:
            kwargs['id'] = [video_id]
        return helix.Videos(self.api, video_ids=None, **kwargs)[0]

    def streams(self, **kwargs) -> 'helix.Streams':
        return helix.Streams(self.api, **kwargs)

    def stream(self, **kwargs) -> 'helix.Stream':
        return self.streams(**kwargs)[0]

    def games(self, **kwargs) -> 'helix.Games':
        return helix.Games(self.api, **kwargs)

    def game(self, **kwargs) -> 'helix.Game':
        return self.games(**kwargs)[0]

    def top_games(self, **kwargs) -> List['helix.Game']:
        return helix.Games(self.api).top(**kwargs)

    def top_game(self) -> 'helix.Game':
        return self.top_games()[0]



from datetime import datetime

from rx.core import typing
from .scheduler import UTC_ZERO
from .virtualtimescheduler import VirtualTimeScheduler


class HistoricalScheduler(VirtualTimeScheduler):
    """Provides a virtual time scheduler that uses datetime for absolute time
    and timedelta for relative time."""

    def __init__(self, initial_clock: datetime = None) -> None:
        """Creates a new historical scheduler with the specified initial clock
        value.

        Args:
            initial_clock: Initial value for the clock.
        """

        super().__init__(initial_clock or UTC_ZERO)

    @property
    def now(self) -> datetime:
        """Represents a notion of time for this scheduler. Tasks being
        scheduled on a scheduler will adhere to the time denoted by this
        property.

        Returns:
             The scheduler's current time, as a datetime instance.
        """

        return self._clock

    @classmethod
    def add(cls,
            absolute: typing.AbsoluteTime,
            relative: typing.RelativeTime
            ) -> typing.AbsoluteTime:
        """Adds a relative time value to an absolute time value.

        Args:
            absolute: Absolute virtual time value.
            relative: Relative virtual time value to add.

        Returns:
            The resulting absolute virtual time sum value.
        """

        return cls.to_datetime(absolute) + cls.to_timedelta(relative)



from typing import List, Tuple, Union, Dict
from collections import namedtuple
from contextlib import contextmanager
from warnings import warn

import rx
from rx.core import Observable
from rx.core.notification import Notification
from rx.scheduler import NewThreadScheduler
from rx.core.typing import Callable, RelativeTime
from rx.testing import TestScheduler, Recorded, ReactiveTest
from rx.core.observable.marbles import parse

new_thread_scheduler = NewThreadScheduler()

MarblesContext = namedtuple('MarblesContext', 'start, cold, hot, exp')


@contextmanager
def marbles_testing(timespan=1.0):
    """
    Initialize a :class:`rx.testing.TestScheduler` and return a namedtuple
    containing the following functions that wrap its methods.

    :func:`cold()`:
    Parse a marbles string and return a cold observable

    :func:`hot()`:
    Parse a marbles string and return a hot observable

    :func:`start()`:
    Start the test scheduler, invoke the create function,
    subscribe to the resulting sequence, dispose the subscription and
    return the resulting records

    :func:`exp()`:
    Parse a marbles string and return a list of records

    Examples:
        >>> with marbles_testing() as (start, cold, hot, exp):
        ...     obs = hot("-a-----b---c-|")
        ...     ex = exp( "-a-----b---c-|")
        ...     results = start(obs)
        ...     assert results == ex

    The underlying test scheduler is initialized with the following
    parameters:
        - created time = 100.0s
        - subscribed = 200.0s
        - disposed = 1000.0s

    **IMPORTANT**: regarding :func:`hot()`, a marble declared as the
    first character will be skipped by the test scheduler.
    E.g. hot("a--b--") will only emit b.
    """

    scheduler = TestScheduler()
    created = 100.0
    disposed = 1000.0
    subscribed = 200.0
    start_called = False
    outside_of_context = False

    def check():
        if outside_of_context:
            warn('context functions should not be called outside of '
                 'with statement.',
                 UserWarning,
                 stacklevel=3,
                 )

        if start_called:
            warn('start() should only be called one time inside '
                 'a with statement.',
                 UserWarning,
                 stacklevel=3,
                 )

    def test_start(create: Union[Observable, Callable[[], Observable]]) -> List[Recorded]:
        nonlocal start_called
        check()

        def default_create():
            return create

        if isinstance(create, Observable):
            create_function = default_create
        else:
            create_function = create

        mock_observer = scheduler.start(
            create=create_function,
            created=created,
            subscribed=subscribed,
            disposed=disposed,
            )
        start_called = True
        return mock_observer.messages

    def test_expected(string: str, lookup: Dict = None, error: Exception = None) -> List[Recorded]:
        messages = parse(
            string,
            timespan=timespan,
            time_shift=subscribed,
            lookup=lookup,
            error=error,
            )
        return messages_to_records(messages)

    def test_cold(string: str, lookup: Dict = None, error: Exception = None) -> Observable:
        check()
        return rx.from_marbles(
            string,
            timespan=timespan,
            lookup=lookup,
            error=error,
            )

    def test_hot(string: str, lookup: Dict = None, error: Exception = None) -> Observable:
        check()
        hot_obs = rx.hot(
            string,
            timespan=timespan,
            duetime=subscribed,
            lookup=lookup,
            error=error,
            scheduler=scheduler,
            )
        return hot_obs

    try:
        yield MarblesContext(test_start, test_cold, test_hot, test_expected)
    finally:
        outside_of_context = True


def messages_to_records(messages: List[Tuple[RelativeTime, Notification]]
                        ) -> List[Recorded]:
    """
    Helper function to convert messages returned by parse() to a list of
    Recorded.
    """
    records = []

    dispatcher = dict(
        N=lambda t, n: ReactiveTest.on_next(t, n.value),
        E=lambda t, n: ReactiveTest.on_error(t, n.exception),
        C=lambda t, n: ReactiveTest.on_completed(t)
        )

    for message in messages:
        time, notification = message
        kind = notification.kind
        record = dispatcher[kind](time, notification)
        records.append(record)

    return records



from typing import Any, Callable

from rx.core import Observable
from rx.core.typing import Predicate, Mapper, RelativeTime
from rx.scheduler import TimeoutScheduler
from rx.disposable import MultipleAssignmentDisposable


def _generate_with_relative_time(initial_state: Any,
                                 condition: Predicate,
                                 iterate: Mapper,
                                 time_mapper: Callable[[Any], RelativeTime]
                                 ) -> Observable:
    """Generates an observable sequence by iterating a state from an
    initial state until the condition fails.

    Example:
        res = source.generate_with_relative_time(0, lambda x: True, lambda x: x + 1, lambda x: 0.5)

    Args:
        initial_state: Initial state.
        condition: Condition to terminate generation (upon returning
            false).
        iterate: Iteration step function.
        time_mapper: Time mapper function to control the speed of
            values being produced each iteration, returning relative times, i.e.
            either floats denoting seconds or instances of timedelta.

    Returns:
        The generated sequence.
    """

    def subscribe(observer, scheduler=None):
        scheduler = scheduler or TimeoutScheduler.singleton()
        mad = MultipleAssignmentDisposable()
        state = initial_state
        has_result = False
        result = None
        first = True
        time = None

        def action(scheduler, _):
            nonlocal state
            nonlocal has_result
            nonlocal result
            nonlocal first
            nonlocal time

            if has_result:
                observer.on_next(result)

            try:
                if first:
                    first = False
                else:
                    state = iterate(state)

                has_result = condition(state)

                if has_result:
                    result = state
                    time = time_mapper(state)

            except Exception as e:  # pylint: disable=broad-except
                observer.on_error(e)
                return

            if has_result:
                mad.disposable = scheduler.schedule_relative(time, action)
            else:
                observer.on_completed()

        mad.disposable = scheduler.schedule_relative(0, action)
        return mad
    return Observable(subscribe)



from typing import Dict, Any, List

import twitch.helix as helix
from twitch.api import API
from .model import Model


class Chatter(Model):

    def __init__(self, api: API, name: str, chatter_type: str):
        super().__init__(api, {})
        self.name: str = name
        self.type: str = chatter_type

    @property
    def user(self) -> 'helix.User':
        source = helix.Helix('')
        source.api = self._api
        source.api.base_url = helix.Helix.BASE_URL
        return source.user(self.name)



import logging
from typing import Callable, Any
from collections import OrderedDict

from rx import operators as ops
from rx.core import Observable
from rx.internal.utils import add_ref
from rx.disposable import SingleAssignmentDisposable, RefCountDisposable, CompositeDisposable
from rx.subject import Subject

log = logging.getLogger("Rx")


def _group_join(right: Observable,
                left_duration_mapper: Callable[[Any], Observable],
                right_duration_mapper: Callable[[Any], Observable],
               ) -> Callable[[Observable], Observable]:
    """Correlates the elements of two sequences based on overlapping
    durations, and groups the results.

    Args:
        right: The right observable sequence to join elements for.
        left_duration_mapper: A function to select the duration (expressed
            as an observable sequence) of each element of the left observable
            sequence, used to determine overlap.
        right_duration_mapper: A function to select the duration (expressed
            as an observable sequence) of each element of the right observable
            sequence, used to determine overlap.

    Returns:
        An observable sequence that contains elements combined into a tuple
    from source elements that have an overlapping duration.
    """

    def nothing(_):
        return None

    def group_join(left: Observable) -> Observable:
        def subscribe(observer, scheduler=None):
            group = CompositeDisposable()
            rcd = RefCountDisposable(group)
            left_map = OrderedDict()
            right_map = OrderedDict()
            left_id = [0]
            right_id = [0]

            def on_next_left(value):
                subject = Subject()

                with left.lock:
                    _id = left_id[0]
                    left_id[0] += 1
                    left_map[_id] = subject

                try:
                    result = (value, add_ref(subject, rcd))
                except Exception as e:
                    log.error("*** Exception: %s" % e)
                    for left_value in left_map.values():
                        left_value.on_error(e)

                    observer.on_error(e)
                    return

                observer.on_next(result)

                for right_value in right_map.values():
                    subject.on_next(right_value)

                md = SingleAssignmentDisposable()
                group.add(md)

                def expire():
                    if _id in left_map:
                        del left_map[_id]
                        subject.on_completed()

                    group.remove(md)

                try:
                    duration = left_duration_mapper(value)
                except Exception as e:
                    for left_value in left_map.values():
                        left_value.on_error(e)

                    observer.on_error(e)
                    return

                def on_error(error):
                    for left_value in left_map.values():
                        left_value.on_error(error)

                    observer.on_error(error)

                md.disposable = duration.pipe(ops.take(1)).subscribe_(nothing, on_error, expire, scheduler)

            def on_error_left(error):
                for left_value in left_map.values():
                    left_value.on_error(error)

                observer.on_error(error)

            group.add(left.subscribe_(on_next_left, on_error_left, observer.on_completed, scheduler))

            def send_right(value):
                with left.lock:
                    _id = right_id[0]
                    right_id[0] += 1
                    right_map[_id] = value

                md = SingleAssignmentDisposable()
                group.add(md)

                def expire():
                    del right_map[_id]
                    group.remove(md)

                try:
                    duration = right_duration_mapper(value)
                except Exception as e:
                    for left_value in left_map.values():
                        left_value.on_error(e)

                    observer.on_error(e)
                    return

                def on_error(error):
                    with left.lock:
                        for left_value in left_map.values():
                            left_value.on_error(error)

                        observer.on_error(error)

                md.disposable = duration.pipe(ops.take(1)).subscribe_(nothing, on_error, expire, scheduler)

                with left.lock:
                    for left_value in left_map.values():
                        left_value.on_next(value)

            def on_error_right(error):
                for left_value in left_map.values():
                    left_value.on_error(error)

                observer.on_error(error)

            group.add(right.subscribe_(send_right, on_error_right, scheduler=scheduler))
            return rcd
        return Observable(subscribe)
    return group_join



from datetime import datetime
from typing import Optional

from rx.scheduler import TimeoutScheduler
from rx.core import Observable, typing
from rx.disposable import MultipleAssignmentDisposable


def observable_timer_date(duetime, scheduler: Optional[typing.Scheduler] = None):
    def subscribe(observer, scheduler_=None):
        _scheduler = scheduler or scheduler_ or TimeoutScheduler.singleton()

        def action(scheduler, state):
            observer.on_next(0)
            observer.on_completed()

        return _scheduler.schedule_absolute(duetime, action)
    return Observable(subscribe)


def observable_timer_duetime_and_period(duetime, period, scheduler: Optional[typing.Scheduler] = None) -> Observable:
    def subscribe(observer, scheduler_=None):
        _scheduler = scheduler or scheduler_ or TimeoutScheduler.singleton()
        nonlocal duetime

        if not isinstance(duetime, datetime):
            duetime = _scheduler.now + _scheduler.to_timedelta(duetime)

        p = max(0.0, _scheduler.to_seconds(period))
        mad = MultipleAssignmentDisposable()
        dt = duetime
        count = 0

        def action(scheduler, state):
            nonlocal dt
            nonlocal count

            if p > 0.0:
                now = scheduler.now
                dt = dt + scheduler.to_timedelta(p)
                if dt <= now:
                    dt = now + scheduler.to_timedelta(p)

            observer.on_next(count)
            count += 1
            mad.disposable = scheduler.schedule_absolute(dt, action)
        mad.disposable = _scheduler.schedule_absolute(dt, action)
        return mad
    return Observable(subscribe)


def observable_timer_timespan(duetime: typing.RelativeTime, scheduler: Optional[typing.Scheduler] = None) -> Observable:
    def subscribe(observer, scheduler_=None):
        _scheduler = scheduler or scheduler_ or TimeoutScheduler.singleton()
        d = _scheduler.to_seconds(duetime)

        def action(scheduler, state):
            observer.on_next(0)
            observer.on_completed()

        if d <= 0.0:
            return _scheduler.schedule(action)
        return _scheduler.schedule_relative(d, action)
    return Observable(subscribe)


def observable_timer_timespan_and_period(duetime: typing.RelativeTime,
                                         period: typing.RelativeTime,
                                         scheduler: Optional[typing.Scheduler] = None
                                         ) -> Observable:
    if duetime == period:
        def subscribe(observer, scheduler_=None):
            _scheduler = scheduler or scheduler_ or TimeoutScheduler.singleton()

            def action(count):
                observer.on_next(count)
                return count + 1

            return _scheduler.schedule_periodic(period, action, state=0)
        return Observable(subscribe)
    return observable_timer_duetime_and_period(duetime, period, scheduler)


def _timer(duetime: typing.AbsoluteOrRelativeTime,
           period: Optional[typing.RelativeTime] = None,
           scheduler: Optional[typing.Scheduler] = None
           ) -> Observable:
    if isinstance(duetime, datetime):
        if period is None:
            return observable_timer_date(duetime, scheduler)
        else:
            return observable_timer_duetime_and_period(duetime, period, scheduler)

    if period is None:
        return observable_timer_timespan(duetime, scheduler)

    return observable_timer_timespan_and_period(duetime, period, scheduler)



from typing import Callable
from rx.core import Observable


def _as_observable() -> Callable[[Observable], Observable]:
    def as_observable(source: Observable) -> Observable:
        """Hides the identity of an observable sequence.

        Args:
            source: Observable source to hide identity from.

        Returns:
            An observable sequence that hides the identity of the
            source sequence.
        """

        def subscribe(observer, scheduler=None):
            return source.subscribe(observer, scheduler=scheduler)

        return Observable(subscribe)
    return as_observable



from typing import Dict, Any

import twitch.helix as helix
import twitch.v5 as v5
from twitch.api import API
from .model import Model


class Video(Model):

    def __init__(self, api: API, data: Dict[str, Any]):
        super().__init__(api, data)

        self.id: str = data.get('id')
        self.user_id: str = data.get('user_id')
        self.user_name: str = data.get('user_name')
        self.title: str = data.get('title')
        self.description: str = data.get('description')
        self.created_at: str = data.get('created_at')
        self.published_at: str = data.get('published_at')
        self.url: str = data.get('url')
        self.thumbnail_url: str = data.get('thumbnail_url')
        self.viewable: str = data.get('viewable')
        self.view_count: int = data.get('view_count')
        self.language: str = data.get('language')
        self.type: str = data.get('type')
        self.duration: str = data.get('duration')

    def __str__(self):
        return self.title

    @property
    def comments(self) -> 'v5.Comments':
        return v5.V5(client_id=self._api.client_id,
                     use_cache=self._api.use_cache,
                     cache_duration=self._api.cache_duration).comments(self.id)

    @property
    def user(self) -> 'helix.User':
        return helix.Users(self._api, int(self.user_id))[0]



import tracemalloc
import unittest
from typing import Dict, Any

import responses

import twitch
from .test_helix_user import TestHelixUser

tracemalloc.start()


class TestHelixVideo(unittest.TestCase):
    video: Dict[str, Dict[str, Any]] = {
        '471855782': {
            'id': '471855782',
            'user_id': '26301881',
            'user_name': 'sodapoppin',
            'title': '2 days til Classic, passing the time til then. ',
            'description': '',
            'created_at': '2019-08-24T21:34:03Z',
            'published_at': '2019-08-24T21:34:03Z',
            'url': 'https://www.twitch.tv/videos/471855782',
            'thumbnail_url': '',
            'viewable': 'public',
            'view_count': 329,
            'language': 'en',
            'type': 'archive',
            'duration': '2h14m55s'
        },
        '471295896': {
            'id': '471295896',
            'user_id': '26301881',
            'user_name': 'sodapoppin',
            'title': '3 days til Classic, passing the time til then. ',
            'description': '',
            'created_at': '2019-08-23T18:40:05Z',
            'published_at': '2019-08-23T18:40:05Z',
            'url': 'https://www.twitch.tv/videos/471295896',
            'thumbnail_url': 'https://static-cdn.jtvnw.net/s3_vods/7d5ae2c2918cf4ca8579_sodapoppin_35403289856_1281380185/thumb/thumb0-%{width}x%{height}.jpg',
            'viewable': 'public',
            'view_count': 5892,
            'language': 'en',
            'type': 'archive',
            'duration': '6h4m13s'}
    }

    def setUp(self) -> None:
        responses.add(responses.GET, 'https://id.twitch.tv/oauth2/token?client_token=id&client_secret=secret&grant_type=client_credentials',
                      match_querystring=True,
                      json={'access_token': 'token'})
        
        responses.add(responses.GET, 'https://api.twitch.tv/helix/videos?id=471855782',
                      match_querystring=True,
                      json={
                          'data': [TestHelixVideo.video['471855782']],
                          'pagination': {'cursor': 'eyJiIjpudWxsLCJhIjp7Ik9mZnNldCI6MX19'}
                      })
        responses.add(responses.GET, 'https://api.twitch.tv/helix/users?login=sodapoppin',
                      match_querystring=True,
                      json={'data': [TestHelixUser.user['sodapoppin']]})

        responses.add(responses.GET, 'https://api.twitch.tv/helix/videos?user_id=26301881&first=2',
                      match_querystring=True,
                      json={
                          'data': [TestHelixVideo.video['471855782'], TestHelixVideo.video['471295896']],
                          'pagination': {'cursor': 'eyJiIjpudWxsLCJhIjp7Ik9mZnNldCI6Mn19'}
                      })

    @responses.activate
    def test_video(self):
        helix = twitch.Helix(client_id='id', bearer_token='token', use_cache=True)

        # Get display name to display name
        self.assertEqual(helix.video('471855782').user_name, 'sodapoppin')

    @responses.activate
    def test_first_videos(self):
        # Should returned cached data from above
        helix = twitch.Helix(client_id='id', bearer_token='token', use_cache=True)

        for video, user_name in zip(helix.user('sodapoppin').videos(first=2), ['sodapoppin', 'sodapoppin']):
            self.assertEqual(video.user_name, user_name)

    @responses.activate
    def test_custom_video_cache(self):
        helix = twitch.Helix(client_id='id', bearer_token='token', use_cache=True)
        _ = helix.video(471855782)

        # Videos have custom caching, such that url should not be cached
        self.assertFalse(helix.api.SHARED_CACHE.has('GET:https://api.twitch.tv/helix/videos?id=471855782'))

        # Cache entries by video id
        self.assertTrue(helix.api.SHARED_CACHE.has('helix.video.471855782'))

        # Flush cache to remove
        helix.api.flush_cache()

        # Check cache flush (cache should be empty)
        self.assertFalse(helix.api.SHARED_CACHE.has('helix.video.471855782'))



from abc import abstractmethod
from .asyncobservable import AsyncObservable


class AsyncObserver(AsyncObservable):
    """An asynchronous observable."""

    __slots__ = ()

    @abstractmethod
    async def on_next_async(self, value):
        return NotImplemented

    @abstractmethod
    async def on_error_async(self, error):
        return NotImplemented

    @abstractmethod
    async def on_completed_async(self):
        return NotImplemented

    async def subscribe_async(self, observer):
        return self



from typing import Callable, Optional
from datetime import timedelta

from rx.core import Observable, typing
from rx.scheduler import TimeoutScheduler
from rx.internal.constants import DELTA_ZERO
from rx.internal.utils import add_ref
from rx.disposable import SingleAssignmentDisposable, CompositeDisposable, RefCountDisposable, SerialDisposable
from rx.subject import Subject


def _window_with_time(timespan: typing.RelativeTime, timeshift: Optional[typing.RelativeTime] = None,
                      scheduler: Optional[typing.Scheduler] = None) -> Callable[[Observable], Observable]:
    if timeshift is None:
        timeshift = timespan

    if not isinstance(timespan, timedelta):
        timespan = timedelta(seconds=timespan)
    if not isinstance(timeshift, timedelta):
        timeshift = timedelta(seconds=timeshift)

    def window_with_time(source: Observable) -> Observable:
        def subscribe(observer, scheduler_=None):
            _scheduler = scheduler or scheduler_ or TimeoutScheduler.singleton()

            timer_d = SerialDisposable()
            next_shift = [timeshift]
            next_span = [timespan]
            total_time = [DELTA_ZERO]
            q = []

            group_disposable = CompositeDisposable(timer_d)
            ref_count_disposable = RefCountDisposable(group_disposable)

            def create_timer():
                m = SingleAssignmentDisposable()
                timer_d.disposable = m
                is_span = False
                is_shift = False

                if next_span[0] == next_shift[0]:
                    is_span = True
                    is_shift = True
                elif next_span[0] < next_shift[0]:
                    is_span = True
                else:
                    is_shift = True

                new_total_time = next_span[0] if is_span else next_shift[0]

                ts = new_total_time - total_time[0]
                total_time[0] = new_total_time
                if is_span:
                    next_span[0] += timeshift

                if is_shift:
                    next_shift[0] += timeshift

                def action(scheduler, state=None):
                    s = None

                    if is_shift:
                        s = Subject()
                        q.append(s)
                        observer.on_next(add_ref(s, ref_count_disposable))

                    if is_span:
                        s = q.pop(0)
                        s.on_completed()

                    create_timer()
                m.disposable = _scheduler.schedule_relative(ts, action)

            q.append(Subject())
            observer.on_next(add_ref(q[0], ref_count_disposable))
            create_timer()

            def on_next(x):
                for s in q:
                    s.on_next(x)

            def on_error(e):
                for s in q:
                    s.on_error(e)

                observer.on_error(e)

            def on_completed():
                for s in q:
                    s.on_completed()

                observer.on_completed()

            group_disposable.add(source.subscribe_(on_next, on_error, on_completed, scheduler_))
            return ref_count_disposable
        return Observable(subscribe)
    return window_with_time



from typing import Optional, List

from twitch.api import API
from twitch.helix.resources.resource import T
from .resource import Resource


class Clips(Resource['helix.Clip']):

    def __init__(self, api: API, **kwargs: Optional):
        super().__init__(api=api, path='users/follows')
        self._kwargs = kwargs

    def _can_paginate(self) -> bool:
        return False

    def _handle_pagination_response(self, response: dict) -> List[T]:
        pass



from typing import Any, Callable, Optional

from .. import typing
from rx.internal import noop, default_error


class Observer(typing.Observer, typing.Disposable):
    """Base class for implementations of the Observer class. This base
    class enforces the grammar of observers where OnError and
    OnCompleted are terminal messages.
    """

    def __init__(self,
                 on_next: Optional[typing.OnNext] = None,
                 on_error: Optional[typing.OnError] = None,
                 on_completed: Optional[typing.OnCompleted] = None
                 ) -> None:
        self.is_stopped = False
        self._handler_on_next = on_next or noop
        self._handler_on_error = on_error or default_error
        self._handler_on_completed = on_completed or noop

    def on_next(self, value: Any) -> None:
        """Notify the observer of a new element in the sequence."""
        if not self.is_stopped:
            self._on_next_core(value)

    def _on_next_core(self, value: Any) -> None:
        """For Subclassing purpose. This method is called by `on_next()`
        method until the observer is stopped.
        """
        self._handler_on_next(value)

    def on_error(self, error: Exception) -> None:
        """Notify the observer that an exception has occurred.

        Args:
            error: The error that occurred.
        """

        if not self.is_stopped:
            self.is_stopped = True
            self._on_error_core(error)

    def _on_error_core(self, error: Exception) -> None:
        """For Subclassing purpose. This method is called by `on_error()`
        method until the observer is stopped.
        """
        self._handler_on_error(error)

    def on_completed(self) -> None:
        """Notifies the observer of the end of the sequence."""

        if not self.is_stopped:
            self.is_stopped = True
            self._on_completed_core()

    def _on_completed_core(self) -> None:
        """For Subclassing purpose. This method is called by `on_completed()`
        method until the observer is stopped.
        """
        self._handler_on_completed()

    def dispose(self) -> None:
        """Disposes the observer, causing it to transition to the
        stopped state."""
        self.is_stopped = True

    def fail(self, exn: Exception) -> bool:
        if not self.is_stopped:
            self.is_stopped = True
            self._on_error_core(exn)
            return True

        return False

    def throw(self, error: Exception) -> None:
        import traceback
        traceback.print_stack()
        if error:
            raise error
        1 / 0  # Raise division by zero

    def to_notifier(self) -> Callable:
        """Creates a notification callback from an observer.

        Returns the action that forwards its input notification to the
        underlying observer."""

        def func(notifier):
            return notifier.accept(self)
        return func

    def as_observer(self) -> 'Observer':
        """Hides the identity of an observer.

        Returns an observer that hides the identity of the specified
        observer.
        """
        return Observer(self.on_next, self.on_error, self.on_completed)



from typing import Callable, Optional
from rx.core import Observable
from rx.core.typing import Mapper, Comparer
from rx.internal.basic import identity, default_comparer


def _distinct_until_changed(
        key_mapper: Optional[Mapper] = None,
        comparer: Optional[Comparer] = None
        ) -> Callable[[Observable], Observable]:

    key_mapper = key_mapper or identity
    comparer = comparer or default_comparer

    def distinct_until_changed(source: Observable) -> Observable:
        """Returns an observable sequence that contains only distinct
        contiguous elements according to the key_mapper and the
        comparer.

        Examples:
            >>> op = distinct_until_changed();
            >>> op = distinct_until_changed(lambda x: x.id)
            >>> op = distinct_until_changed(lambda x: x.id, lambda x, y: x == y)

        Args:
            key_mapper: [Optional] A function to compute the comparison
                key for each element. If not provided, it projects the
                value.
            comparer: [Optional] Equality comparer for computed key
                values. If not provided, defaults to an equality
                comparer function.

        Returns:
            An observable sequence only containing the distinct
            contiguous elements, based on a computed key value, from
            the source sequence.
        """
        def subscribe(observer, scheduler=None):
            has_current_key = [False]
            current_key = [None]

            def on_next(value):
                comparer_equals = False
                try:
                    key = key_mapper(value)
                except Exception as exception:  # pylint: disable=broad-except
                    observer.on_error(exception)
                    return

                if has_current_key[0]:
                    try:
                        comparer_equals = comparer(current_key[0], key)
                    except Exception as exception:  # pylint: disable=broad-except
                        observer.on_error(exception)
                        return

                if not has_current_key[0] or not comparer_equals:
                    has_current_key[0] = True
                    current_key[0] = key
                    observer.on_next(value)

            return source.subscribe_(on_next, observer.on_error, observer.on_completed, scheduler=scheduler)
        return Observable(subscribe)
    return distinct_until_changed



from typing import List, Callable

from twitch.helix.helix import Helix
from twitch.helix.models import Clip, Follow, Game, Stream, User, Video
from twitch.helix.resources import Clips, Follows, Games, Streams, StreamNotFound, Users, Videos

__all__: List[Callable] = [
    Helix,
    Stream, StreamNotFound, Streams,
    Video, Videos,
    User, Users,
    Follow, Follows,
    Game, Games,
]



from typing import List

from rx.disposable import Disposable
from rx.core import Observable, typing
from rx.scheduler import VirtualTimeScheduler

from .recorded import Recorded
from .subscription import Subscription


class HotObservable(Observable):
    def __init__(self, scheduler: VirtualTimeScheduler, messages: List[Recorded]) -> None:
        super().__init__()

        self.scheduler: VirtualTimeScheduler = scheduler
        self.messages = messages
        self.subscriptions: List[Subscription] = []
        self.observers: List[typing.Observer] = []

        observable = self

        def get_action(notification):
            def action(scheduler, state):
                for observer in observable.observers[:]:
                    notification.accept(observer)
                return Disposable()
            return action

        for message in self.messages:
            notification = message.value

            # Warning: Don't make closures within a loop
            action = get_action(notification)
            scheduler.schedule_absolute(message.time, action)

    def _subscribe_core(self, observer=None, scheduler=None) -> typing.Disposable:
        self.observers.append(observer)
        self.subscriptions.append(Subscription(self.scheduler.clock))
        index = len(self.subscriptions) - 1

        def dispose_action():
            self.observers.remove(observer)
            start = self.subscriptions[index].subscribe
            end = self.scheduler.clock
            self.subscriptions[index] = Subscription(start, end)

        return Disposable(dispose_action)



from typing import Any, List

from rx.core.typing import Observer
from rx.core.notification import OnNext, OnError, OnCompleted
from rx.scheduler import VirtualTimeScheduler

from .recorded import Recorded


class MockObserver(Observer):

    def __init__(self, scheduler: VirtualTimeScheduler) -> None:
        self.scheduler: VirtualTimeScheduler = scheduler
        self.messages: List[Recorded] = []

    def on_next(self, value: Any) -> None:
        self.messages.append(Recorded(self.scheduler.clock, OnNext(value)))

    def on_error(self, error: Exception) -> None:
        self.messages.append(Recorded(self.scheduler.clock, OnError(error)))

    def on_completed(self) -> None:
        self.messages.append(Recorded(self.scheduler.clock, OnCompleted()))



from typing import Optional

from twitch.helix import User, Helix
from .chat import Chat


class Message:

    def __init__(self,
                 channel: str,
                 sender: str,
                 text: str,
                 helix_api: Optional[Helix] = None,
                 chat: Optional[Chat] = None):
        self.channel: str = channel
        self.sender: str = sender
        self.text: str = text
        self.helix: Optional[Helix] = helix_api
        self.chat: Optional[Chat] = chat

    @property
    def user(self) -> Optional[User]:
        return self.helix.user(self.sender) if self.helix else None



from typing import Optional

from rx.core import Observable, typing
from rx.disposable import CompositeDisposable, SingleAssignmentDisposable


def _combine_latest(*sources: Observable) -> Observable:
    """Merges the specified observable sequences into one observable
    sequence by creating a tuple whenever any of the
    observable sequences produces an element.

    Examples:
        >>> obs = combine_latest(obs1, obs2, obs3)

    Returns:
        An observable sequence containing the result of combining
        elements of the sources into a tuple.
    """

    parent = sources[0]

    def subscribe(observer: typing.Observer,
                  scheduler: Optional[typing.Scheduler] = None
                  ) -> CompositeDisposable:

        n = len(sources)
        has_value = [False] * n
        has_value_all = [False]
        is_done = [False] * n
        values = [None] * n

        def _next(i):
            has_value[i] = True

            if has_value_all[0] or all(has_value):
                res = tuple(values)
                observer.on_next(res)

            elif all([x for j, x in enumerate(is_done) if j != i]):
                observer.on_completed()

            has_value_all[0] = all(has_value)

        def done(i):
            is_done[i] = True
            if all(is_done):
                observer.on_completed()

        subscriptions = [None] * n

        def func(i):
            subscriptions[i] = SingleAssignmentDisposable()

            def on_next(x):
                with parent.lock:
                    values[i] = x
                    _next(i)

            def on_completed():
                with parent.lock:
                    done(i)

            subscriptions[i].disposable = sources[i].subscribe_(on_next, observer.on_error, on_completed, scheduler)

        for idx in range(n):
            func(idx)
        return CompositeDisposable(subscriptions)
    return Observable(subscribe)



from typing import List, Union, Tuple, Generator

import twitch.helix as helix
import twitch.v5 as v5
from twitch.api import API
from .resource import Resource


class VideosAPIException(Exception):
    pass


class Videos(Resource['helix.Video']):
    DEFAULT_FIRST: int = 20
    FIRST_API_LIMIT: int = 100
    ID_API_LIMIT: int = 100
    CACHE_PREFIX: str = 'helix.video.'

    def __init__(self,
                 api: API,
                 video_ids: Union[int, List[int]] = None,
                 **kwargs):
        super().__init__(api=api, path='videos')

        # Store kwargs as class property for __iter__
        self._kwargs = kwargs

        # 'id' parameter can be a singular or a list
        # Create list of video ids by combining video_ids and kwargs['id']

        # Convert singular string to list
        if 'id' in self._kwargs and type(self._kwargs['id']) == str:
            self._kwargs['id'] = [self._kwargs['id']]

        self._kwargs['id'] = list(self._kwargs['id']) if 'id' in self._kwargs.keys() else []
        self._kwargs['id'] = self._kwargs['id'] + list(video_ids) if video_ids else self._kwargs['id']

        # Convert to integers
        self._kwargs['id'] = [int(x) for x in self._kwargs['id']]

        # Remove duplicates
        self._kwargs['id'] = list(set(self._kwargs['id'])) if self._kwargs['id'] else []

        # Download video ids
        if len(self._kwargs['id']) > 0:
            self._download_video_ids()

    def _can_paginate(self) -> bool:
        """
        Kwargs must include user_id or game_id
        :return: If resource can paginate
        """
        # todo: maybe raise VideosAPIException('A user_id or a game_id must be specified.')
        return len([key for key in self._kwargs.keys() if key in ['user_id', 'game_id']]) == 1

    def _cache_videos(self, videos: List['helix.Video']) -> None:
        """
        Custom video cache
        Cache individual videos
        :param videos: Helix videos
        :return: None
        """
        if self._api.use_cache:
            for video in videos:
                API.SHARED_CACHE.set(f'{Videos.CACHE_PREFIX}{video.id}', video.data)

    def _handle_pagination_response(self, response: dict) -> List['helix.Video']:
        """
        Custom handling for video pagination
        :param response: API response data
        :return: Videos
        """
        videos: List['helix.Video'] = [helix.Video(api=self._api, data=video) for video in response['data']]
        self._cache_videos(videos)

        return videos

    def _next_videos_page(self, ignore_cache: bool = False) -> List['helix.Video']:
        """
        Video pagination
        :param ignore_cache: Ignore API cache
        :return: Videos
        """
        response: dict = self._next_page(ignore_cache=ignore_cache)

        return self._handle_pagination_response(response)

    def _cache_download(self, video_ids: List[int]) -> List[int]:
        """
        Fetch data from cache
        :param video_ids: Lookup the video ids
        :return: Cache hits (video ids)
        """
        cache_hits: list = []
        for video_id in video_ids:
            cache_data: dict = API.SHARED_CACHE.get(f'{Videos.CACHE_PREFIX}{video_id}')
            if cache_data:
                self._data.append(helix.Video(api=self._api, data=cache_data))
                cache_hits.append(video_id)

        return cache_hits

    def _download_video_ids(self) -> None:
        """
        Download videos by list of video IDs
        :return:
        """
        # Custom cache lookup
        if self._api.use_cache:
            cache_hits: List[int] = self._cache_download(self._kwargs['id'])

            # Removed cached ids from kwargs
            self._kwargs['id'] = [n for n in self._kwargs['id'] if n not in cache_hits]

        # Download uncached videos from API
        if len(self._kwargs['id']) > 0:

            # When the number of IDs exceeds API limitations, divide into multiple requests
            remaining_video_ids: list = self._kwargs['id']

            while remaining_video_ids:
                self._kwargs['id'] = remaining_video_ids[:Videos.ID_API_LIMIT]

                # Ignore default caching method, as we want to cache individual videos and not a collection of videos.
                videos: List[helix.Video] = self._next_videos_page(ignore_cache=True)

                # Save videos
                self._data.extend(videos)

                # Update remaining video ids
                remaining_video_ids = [] if len(videos) < len(remaining_video_ids) else remaining_video_ids[
                                                                                        Videos.ID_API_LIMIT:]

    @property
    def comments(self) -> Generator[Tuple['helix.Video', 'v5.Comments'], None, None]:
        for video in self:
            yield video, video.comments



from functools import update_wrapper
from types import FunctionType
from typing import cast, Any, Callable, Iterable

from rx.disposable import CompositeDisposable


def add_ref(xs, r):
    from rx.core import Observable

    def subscribe(observer, scheduler=None):
        return CompositeDisposable(r.disposable, xs.subscribe(observer))

    return Observable(subscribe)


def is_future(fut: Any) -> bool:
    return callable(getattr(fut, 'add_done_callback', None))


def infinite() -> Iterable[int]:
    n = 0
    while True:
        yield n
        n += 1


def alias(name: str, doc: str, fun: Callable[..., Any]) -> Callable[..., Any]:
    # Adapted from https://stackoverflow.com/questions/13503079/how-to-create-a-copy-of-a-python-function#
    # See also help(type(lambda: 0))
    _fun = cast(FunctionType, fun)
    args = (_fun.__code__, _fun.__globals__)
    kwargs = {
        'name': name,
        'argdefs': _fun.__defaults__,
        'closure': _fun.__closure__
    }
    alias = FunctionType(*args, **kwargs)  # type: ignore
    alias = cast(FunctionType, update_wrapper(alias, _fun))
    alias.__kwdefaults__ = _fun.__kwdefaults__
    alias.__doc__ = doc
    return alias


class NotSet:
    """Sentinel value."""

    def __eq__(self, other):
        return self is other

    def __repr__(self):
        return 'NotSet'



from typing import Callable

from rx import operators as ops
from rx.core import Observable, pipe
from rx.core.typing import Predicate


def _all(predicate: Predicate) -> Callable[[Observable], Observable]:

    filtering = ops.filter(lambda v: not predicate(v))
    mapping = ops.map(lambda b: not b)
    some = ops.some()

    return pipe(
        filtering,
        some,
        mapping
    )



from typing import Callable, Optional

from rx import operators as ops
from rx.core import Observable, pipe
from rx.core.typing import Predicate


def _single(predicate: Optional[Predicate] = None) -> Callable[[Observable], Observable]:
    """Returns the only element of an observable sequence that satisfies the
    condition in the optional predicate, and reports an exception if there
    is not exactly one element in the observable sequence.

    Example:
        >>> res = single()
        >>> res = single(lambda x: x == 42)

    Args:
        predicate -- [Optional] A predicate function to evaluate for
            elements in the source sequence.

    Returns:
        An observable sequence containing the single element in the
        observable sequence that satisfies the condition in the predicate.
    """

    if predicate:
        return pipe(ops.filter(predicate), ops.single())
    else:
        return ops.single_or_default_async(False)



from rx.disposable import CompositeDisposable, SingleAssignmentDisposable

from rx.core import Observable
from rx.internal.utils import NotSet


def _with_latest_from(parent: Observable, *sources: Observable) -> Observable:
    NO_VALUE = NotSet()

    def subscribe(observer, scheduler=None):
        def subscribe_all(parent, *children):

            values = [NO_VALUE for _ in children]

            def subscribe_child(i, child):
                subscription = SingleAssignmentDisposable()

                def on_next(value):
                    with parent.lock:
                        values[i] = value
                subscription.disposable = child.subscribe_(on_next, observer.on_error, scheduler=scheduler)
                return subscription

            parent_subscription = SingleAssignmentDisposable()

            def on_next(value):
                with parent.lock:
                    if NO_VALUE not in values:
                        result = (value,) + tuple(values)
                        observer.on_next(result)

            disp = parent.subscribe_(on_next, observer.on_error, observer.on_completed, scheduler)
            parent_subscription.disposable = disp

            children_subscription = [subscribe_child(i, child) for i, child in enumerate(children)]

            return [parent_subscription] + children_subscription
        return CompositeDisposable(subscribe_all(parent, *sources))
    return Observable(subscribe)



from typing import Callable

import rx
from rx.core import Observable


def _on_error_resume_next(second: Observable) -> Callable[[Observable], Observable]:
    def on_error_resume_next(source: Observable) -> Observable:
        return rx.on_error_resume_next(source, second)
    return on_error_resume_next



from typing import Optional

from rx import timer
from rx.core import Observable, typing


def _interval(period: typing.RelativeTime,
              scheduler: Optional[typing.Scheduler] = None
              ) -> Observable:

    return timer(period, period, scheduler)



from typing import List, Callable

from .clips import Clips
from .follows import Follows
from .games import Games
from .resource import Resource
from .streams import Streams, StreamNotFound
from .users import Users
from .videos import Videos

__all__: List[Callable] = [
    Follows,
    Games,
    Resource,
    Streams,
    StreamNotFound,
    Users,
    Videos,
]



from typing import Any, Optional

from rx.core import typing, Observable
from rx.scheduler import ImmediateScheduler


def _empty(scheduler: Optional[typing.Scheduler] = None) -> Observable:
    def subscribe(observer: typing.Observer,
                  scheduler_: Optional[typing.Scheduler] = None
                  ) -> typing.Disposable:

        _scheduler = scheduler or scheduler_ or ImmediateScheduler.singleton()

        def action(_: typing.Scheduler, __: Any) -> None:
            observer.on_completed()

        return _scheduler.schedule(action)
    return Observable(subscribe)



from .gtkscheduler import GtkScheduler
from .tkinterscheduler import TkinterScheduler
from .pygamescheduler import PyGameScheduler
from .qtscheduler import QtScheduler
from .wxscheduler import WxScheduler



import inspect

from matplotlib import _api


class Substitution:
    """
    A decorator that performs %-substitution on an object's docstring.

    This decorator should be robust even if ``obj.__doc__`` is None (for
    example, if -OO was passed to the interpreter).

    Usage: construct a docstring.Substitution with a sequence or dictionary
    suitable for performing substitution; then decorate a suitable function
    with the constructed object, e.g.::

        sub_author_name = Substitution(author='Jason')

        @sub_author_name
        def some_function(x):
            "%(author)s wrote this function"

        # note that some_function.__doc__ is now "Jason wrote this function"

    One can also use positional arguments::

        sub_first_last_names = Substitution('Edgar Allen', 'Poe')

        @sub_first_last_names
        def some_function(x):
            "%s %s wrote the Raven"
    """
    def __init__(self, *args, **kwargs):
        if args and kwargs:
            raise TypeError("Only positional or keyword args are allowed")
        self.params = args or kwargs

    def __call__(self, func):
        if func.__doc__:
            func.__doc__ = inspect.cleandoc(func.__doc__) % self.params
        return func

    def update(self, *args, **kwargs):
        """
        Update ``self.params`` (which must be a dict) with the supplied args.
        """
        self.params.update(*args, **kwargs)

    @classmethod
    @_api.deprecated("3.3", alternative="assign to the params attribute")
    def from_params(cls, params):
        """
        In the case where the params is a mutable sequence (list or
        dictionary) and it may change before this class is called, one may
        explicitly use a reference to the params rather than using *args or
        **kwargs which will copy the values and not reference them.

        :meta private:
        """
        result = cls()
        result.params = params
        return result


def copy(source):
    """Copy a docstring from another source function (if present)."""
    def do_copy(target):
        if source.__doc__:
            target.__doc__ = source.__doc__
        return target
    return do_copy


# Create a decorator that will house the various docstring snippets reused
# throughout Matplotlib.
interpd = Substitution()
dedent_interpd = interpd



from typing import Callable, Any
from collections import OrderedDict

from rx.operators import take
from rx.core import Observable
from rx.internal import noop
from rx.disposable import SingleAssignmentDisposable, CompositeDisposable


def _join(right: Observable,
          left_duration_mapper: Callable[[Any], Observable],
          right_duration_mapper: Callable[[Any], Observable],
          ) -> Callable[[Observable], Observable]:

    def join(source: Observable) -> Observable:
        """Correlates the elements of two sequences based on
        overlapping durations.

        Args:
            source: Source observable.

        Return:
            An observable sequence that contains elements
            combined into a tuple from source elements that have an overlapping
            duration.
        """

        left = source

        def subscribe(observer, scheduler=None):
            group = CompositeDisposable()
            left_done = False
            left_map = OrderedDict()
            left_id = 0
            right_done = False
            right_map = OrderedDict()
            right_id = 0

            def on_next_left(value):
                nonlocal left_id
                duration = None
                current_id = left_id
                left_id += 1
                md = SingleAssignmentDisposable()

                left_map[current_id] = value
                group.add(md)

                def expire():
                    if current_id in left_map:
                        del left_map[current_id]
                    if not len(left_map) and left_done:
                        observer.on_completed()

                    return group.remove(md)

                try:
                    duration = left_duration_mapper(value)
                except Exception as exception:
                    observer.on_error(exception)
                    return

                md.disposable = duration.pipe(take(1)).subscribe_(noop, observer.on_error, lambda: expire(), scheduler)

                for val in right_map.values():
                    result = (value, val)
                    observer.on_next(result)

            def on_completed_left():
                nonlocal left_done
                left_done = True
                if right_done or not len(left_map):
                    observer.on_completed()

            group.add(left.subscribe_(on_next_left, observer.on_error, on_completed_left, scheduler))

            def on_next_right(value):
                nonlocal right_id
                duration = None
                current_id = right_id
                right_id += 1
                md = SingleAssignmentDisposable()
                right_map[current_id] = value
                group.add(md)

                def expire():
                    if current_id in right_map:
                        del right_map[current_id]
                    if not len(right_map) and right_done:
                        observer.on_completed()

                    return group.remove(md)

                try:
                    duration = right_duration_mapper(value)
                except Exception as exception:
                    observer.on_error(exception)
                    return

                md.disposable = duration.pipe(take(1)).subscribe_(noop, observer.on_error, lambda: expire(), scheduler)

                for val in left_map.values():
                    result = (val, value)
                    observer.on_next(result)

            def on_completed_right():
                nonlocal right_done
                right_done = True
                if left_done or not len(right_map):
                    observer.on_completed()

            group.add(right.subscribe_(on_next_right, observer.on_error, on_completed_right, scheduler))
            return group
        return Observable(subscribe)
    return join



from typing import Callable, Optional
from datetime import datetime

from rx.core import Observable, typing
from rx.disposable import CompositeDisposable
from rx.scheduler import TimeoutScheduler


def _take_until_with_time(end_time: typing.AbsoluteOrRelativeTime, scheduler: Optional[typing.Scheduler] = None
                          ) -> Callable[[Observable], Observable]:
    def take_until_with_time(source: Observable) -> Observable:
        """Takes elements for the specified duration until the specified end
        time, using the specified scheduler to run timers.

        Examples:
            >>> res = take_until_with_time(source)

        Args:
            source: Source observale to take elements from.

        Returns:
            An observable sequence with the elements taken
            until the specified end time.
        """

        def subscribe(observer, scheduler_=None):
            _scheduler = scheduler or scheduler_ or TimeoutScheduler.singleton()

            if isinstance(end_time, datetime):
                scheduler_method = _scheduler.schedule_absolute
            else:
                scheduler_method = _scheduler.schedule_relative

            def action(scheduler, state):
                observer.on_completed()

            task = scheduler_method(end_time, action)
            return CompositeDisposable(task, source.subscribe(observer, scheduler=scheduler_))
        return Observable(subscribe)
    return take_until_with_time



from typing import Callable, Union, cast
from asyncio import Future

import rx
from rx.core import Observable
from rx.core.typing import Scheduler
from rx.internal.utils import is_future


def _if_then(condition: Callable[[], bool],
             then_source: Union[Observable, Future],
             else_source: Union[None, Observable, Future] = None
             ) -> Observable:
    """Determines whether an observable collection contains values.

    Example:
    1 - res = rx.if_then(condition, obs1)
    2 - res = rx.if_then(condition, obs1, obs2)

    Args:
        condition: The condition which determines if the then_source or
            else_source will be run.
        then_source: The observable sequence or Promise that
            will be run if the condition function returns true.
        else_source: [Optional] The observable sequence or
            Promise that will be run if the condition function returns
            False. If this is not provided, it defaults to
            rx.empty

    Returns:
        An observable sequence which is either the then_source or
        else_source.
    """

    else_source = else_source or rx.empty()

    then_source = rx.from_future(cast(Future, then_source)) if is_future(then_source) else then_source
    else_source = rx.from_future(cast(Future, else_source)) if is_future(else_source) else else_source

    def factory(_: Scheduler):
        return then_source if condition() else else_source

    return rx.defer(factory)



from typing import Callable, Optional
from collections import OrderedDict

from rx import operators as ops
from rx.core import Observable, GroupedObservable
from rx.core.typing import Mapper
from rx.subject import Subject
from rx.disposable import CompositeDisposable, RefCountDisposable, SingleAssignmentDisposable
from rx.internal.basic import identity


def _group_by_until(key_mapper: Mapper,
                    element_mapper: Optional[Mapper],
                    duration_mapper: Callable[[GroupedObservable], Observable],
                    subject_mapper: Optional[Callable[[], Subject]] = None,
                    ) -> Callable[[Observable], Observable]:
    """Groups the elements of an observable sequence according to a
    specified key mapper function. A duration mapper function is used
    to control the lifetime of groups. When a group expires, it receives
    an OnCompleted notification. When a new element with the same key
    value as a reclaimed group occurs, the group will be reborn with a
    new lifetime request.

    Examples:
        >>> group_by_until(lambda x: x.id, None, lambda : rx.never())
        >>> group_by_until(lambda x: x.id,lambda x: x.name, lambda grp: rx.never())
        >>> group_by_until(lambda x: x.id, lambda x: x.name, lambda grp: rx.never(), lambda: ReplaySubject())

    Args:
        key_mapper: A function to extract the key for each element.
        duration_mapper: A function to signal the expiration of a group.
        subject_mapper: A function that returns a subject used to initiate
            a grouped observable. Default mapper returns a Subject object.

    Returns: a sequence of observable groups, each of which corresponds to
    a unique key value, containing all elements that share that same key
    value. If a group's lifetime expires, a new group with the same key
    value can be created once an element with such a key value is
    encountered.
    """

    element_mapper = element_mapper or identity

    default_subject_mapper = Subject
    subject_mapper = subject_mapper or default_subject_mapper

    def group_by_until(source: Observable) -> Observable:
        def subscribe(observer, scheduler=None):
            writers = OrderedDict()
            group_disposable = CompositeDisposable()
            ref_count_disposable = RefCountDisposable(group_disposable)

            def on_next(x):
                writer = None
                key = None

                try:
                    key = key_mapper(x)
                except Exception as e:
                    for wrt in writers.values():
                        wrt.on_error(e)

                    observer.on_error(e)
                    return

                fire_new_map_entry = False
                writer = writers.get(key)
                if not writer:
                    try:
                        writer = subject_mapper()
                    except Exception as e:
                        for wrt in writers.values():
                            wrt.on_error(e)

                        observer.on_error(e)
                        return

                    writers[key] = writer
                    fire_new_map_entry = True

                if fire_new_map_entry:
                    group = GroupedObservable(key, writer, ref_count_disposable)
                    duration_group = GroupedObservable(key, writer)
                    try:
                        duration = duration_mapper(duration_group)
                    except Exception as e:
                        for wrt in writers.values():
                            wrt.on_error(e)

                        observer.on_error(e)
                        return

                    observer.on_next(group)
                    sad = SingleAssignmentDisposable()
                    group_disposable.add(sad)

                    def expire():
                        if writers[key]:
                            del writers[key]
                            writer.on_completed()

                        group_disposable.remove(sad)

                    def on_next(value):
                        pass

                    def on_error(exn):
                        for wrt in writers.values():
                            wrt.on_error(exn)
                        observer.on_error(exn)

                    def on_completed():
                        expire()

                    sad.disposable = duration.pipe(ops.take(1)).subscribe_(on_next, on_error, on_completed, scheduler)

                try:
                    element = element_mapper(x)
                except Exception as error:
                    for wrt in writers.values():
                        wrt.on_error(error)

                    observer.on_error(error)
                    return

                writer.on_next(element)

            def on_error(ex):
                for wrt in writers.values():
                    wrt.on_error(ex)

                observer.on_error(ex)

            def on_completed():
                for wrt in writers.values():
                    wrt.on_completed()

                observer.on_completed()

            group_disposable.add(source.subscribe_(on_next, on_error, on_completed, scheduler))
            return ref_count_disposable
        return Observable(subscribe)
    return group_by_until



from sys import maxsize
from typing import Callable, Optional

from rx import operators as ops
from rx.core import Observable


# pylint: disable=redefined-builtin
def _slice(start: Optional[int] = None,
           stop: Optional[int] = None,
           step: Optional[int] = None
           ) -> Callable[[Observable], Observable]:
    _start: int = 0 if start is None else start
    _stop: int = maxsize if stop is None else stop
    _step: int = 1 if step is None else step

    pipeline = []

    def slice(source: Observable) -> Observable:
        """The partially applied slice operator.

        Slices the given observable. It is basically a wrapper around the operators
        :func:`skip <rx.operators.skip>`,
        :func:`skip_last <rx.operators.skip_last>`,
        :func:`take <rx.operators.take>`,
        :func:`take_last <rx.operators.take_last>` and
        :func:`filter <rx.operators.filter>`.

        The following diagram helps you remember how slices works with streams.

        Positive numbers are relative to the start of the events, while negative
        numbers are relative to the end (close) of the stream.

        .. code::

            r---e---a---c---t---i---v---e---!
            0   1   2   3   4   5   6   7   8
           -8  -7  -6  -5  -4  -3  -2  -1   0

        Examples:
            >>> result = source.slice(1, 10)
            >>> result = source.slice(1, -2)
            >>> result = source.slice(1, -1, 2)

        Args:
            source: Observable to slice

        Returns:
            A sliced observable sequence.
        """

        if _stop >= 0:
            pipeline.append(ops.take(_stop))

        if _start > 0:
            pipeline.append(ops.skip(_start))
        elif _start < 0:
            pipeline.append(ops.take_last(-_start))

        if _stop < 0:
            pipeline.append(ops.skip_last(-_stop))

        if _step > 1:
            pipeline.append(ops.filter_indexed(lambda x, i: i % _step == 0))
        elif _step < 0:
            # Reversing events is not supported
            raise TypeError('Negative step not supported.')

        return source.pipe(*pipeline)
    return slice
